/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/* ts30_42S src/avs/fs/mmfs/samples/util/tsinode.c 1.4                    */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2002               */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
static char sccsid[] = "@(#)73	1.30.1.2  src/avs/fs/mmfs/samples/util/tsinode.c, mmfs, avs_rhrz, rhrz1321a 10/10/12 11:23:06";

/*====================================================================
 *
 * tsinode command: list all inodes.
 *
 * Syntax: tsinode pathname [startinode [endinode]]
 *
 *   The "pathname" parameter should be the directory where the root of
 *   the filesystem/snapshot is found.
 *
 *==================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <libgen.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#define GPFS_64BIT_INODES
#include <gpfs.h>

#define INODENUM_MAX 0x7FFFFFFFFFFFFFFEULL /* 2^63-2 */
#define MAX_XATTRS 16

static int headerPrinted = 0;
static char PrintHex[16] =
  { '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
#define PRINTHEX(x) PrintHex[x]

/* print usage message and exit */
static void usage(char *argv0)
{
  fprintf(stderr, "\
Usage: %s pathname [-x [xattr1 xattr2 ...]][startinode [endinode]][-c]\n\
                        [--fset][--fsname fsName][--fsetname filesetName]\n\
                        [--snapname snapshotName][--prevsnapname snapshotName]\n", basename(argv0));
  exit(1);
}

char *
pmode(mode_t aflag, char *buf)
{
  /* these arrays are declared static to allow initializations */
  static int      m0[] = { 1, S_IREAD>>0, 'r', '-' };
  static int      m1[] = { 1, S_IWRITE>>0, 'w', '-' };
  static int      m2[] = { 3, S_ISUID|S_IEXEC, 's', S_IEXEC, 'x', S_ISUID, 'S', '-' };
  static int      m3[] = { 1, S_IREAD>>3, 'r', '-' };
  static int      m4[] = { 1, S_IWRITE>>3, 'w', '-' };
  static int      m5[] = { 3, S_ISGID|(S_IEXEC>>3), 's',
                              S_IEXEC>>3, 'x', S_ISGID, 'S', '-' };
  static int      m6[] = { 1, S_IREAD>>6, 'r', '-' };
  static int      m7[] = { 1, S_IWRITE>>6, 'w', '-' };
  static int      m8[] = { 3, S_ISVTX|(S_IEXEC>>6), 't', S_IEXEC>>6, 'x', S_ISVTX, 'T', '-' };
  static int  *m[] = { m0, m1, m2, m3, m4, m5, m6, m7, m8 };

  int i, j, n;
  int *p;

  buf[0] = S_ISREG(aflag)?  '-':
    S_ISDIR(aflag)?  'd':
    S_ISLNK(aflag)?  'l':
    S_ISFIFO(aflag)? 'p':
    S_ISCHR(aflag)?  'c':
    S_ISBLK(aflag)?  'b':
    S_ISSOCK(aflag)? 's': '?';
  for (i = 0; i <= 8; i++)
  {
    for (n = m[i][0], j = 1; n > 0; n--, j+=2)
    {
      p = m[i];
      if ((aflag & p[j]) == p[j])
      {
        j++;
        break;
      }
    }
    buf[i+1] = p[j];
  }
  buf[10] = '\0';
  return buf;
}

#define DevMajor(x) ((unsigned)(x)>>16)
#define DevMinor(x) ((x)&0xFFFF)

static void
print_vattr(gpfs_iscan_t *iscanP,
            gpfs_snapid_t snapId,
            const gpfs_iattr_t *iattrP,
            int printCompare)
{
  int rc;
  long long space;
  char mbuf[32];
  char buf[4096];
  char nbuf[GPFS_MAXNAMLEN+1];
  char *bufP = buf;
  unsigned long long ino, gen, uid, gid, nlink;
  long long asec, msec, csec, crsec;
  long ansec;

  if (!headerPrinted)
  {
    headerPrinted = 1;
    printf("              inode       gen    uid    gid       size      space       mode     nlink          atime                mtime               ctime        snapId flags\n");
  }

  if (iattrP->ia_flags & GPFS_IAFLAG_ERROR)
  {
    printf("%9d Invalid", iattrP->ia_inode);
    return;
  }
  
  /* Convert inode fields to common format */
  ino = iattrP->ia_inode;
  gen = iattrP->ia_gen;
  uid = iattrP->ia_uid;
  gid = iattrP->ia_gid;
  nlink = iattrP->ia_nlink;
  asec = iattrP->ia_atime.tv_sec;
  ansec = iattrP->ia_atime.tv_nsec;
  msec = iattrP->ia_mtime.tv_sec;
  csec = iattrP->ia_ctime.tv_sec;

  /* atime allowed to change between snapshots */
  if (printCompare > 0)
  {
    asec = 0;
    ansec = 0;
  }

  /* Dump contents of Inode */
  bufP += sprintf(bufP, "%19lld %9lld %6lld %6lld %10lld",
                  ino, gen, uid, gid, iattrP->ia_size);
  space = iattrP->ia_blocks * 512;
  bufP += sprintf(bufP, " %10lld %s", space, pmode(iattrP->ia_mode, mbuf));

  bufP += sprintf(bufP, " %9lld %9lld.%09d %9lld.%09d %9lld.%09d",
                  nlink,
                  asec, ansec,
                  msec, iattrP->ia_mtime.tv_nsec,
                  csec, iattrP->ia_ctime.tv_nsec);

  bufP += sprintf(bufP, " %6u", (int)snapId);

  if (S_ISBLK(iattrP->ia_mode)  ||  S_ISCHR(iattrP->ia_mode))
    bufP += sprintf(bufP, " rdev=%d,%d",
                    DevMajor(iattrP->ia_rdev),
                    DevMinor(iattrP->ia_rdev));

  if (iattrP->ia_blocksize != 262144)
    bufP += sprintf(bufP, " blk=%d", iattrP->ia_blocksize);

  /* Can't compare on the file's mod snap id */
  if (printCompare == 0)
    if (iattrP->ia_modsnapid != 0)
      bufP += sprintf(bufP, " modsnap=%d", iattrP->ia_modsnapid);
  
  if (iattrP->ia_xperm)
  {
    bufP += sprintf(bufP, " extperms=0x%X", iattrP->ia_xperm);
    if (iattrP->ia_xperm & GPFS_IAXPERM_ACL)
      bufP += sprintf(bufP, ",acl");
    if (iattrP->ia_xperm & GPFS_IAXPERM_XATTR)
      bufP += sprintf(bufP, ",xa");
    if (iattrP->ia_xperm & GPFS_IAXPERM_DMATTR)
      bufP += sprintf(bufP, ",dm");
    if (iattrP->ia_xperm & GPFS_IAXPERM_DOSATTR)
      bufP += sprintf(bufP, ",dos");
    if (iattrP->ia_xperm & GPFS_IAXPERM_RPATTR)
      bufP += sprintf(bufP, ",rp");
  }

  if (iattrP->ia_flags & GPFS_IAFLAG_USRQUOTA)
    bufP += sprintf(bufP, " userquota");
  if (iattrP->ia_flags & GPFS_IAFLAG_GRPQUOTA)
    bufP += sprintf(bufP, " groupquota");

  if (iattrP->ia_flags & GPFS_IAFLAG_REPLMETA)
    bufP += sprintf(bufP, " replmeta");
  if (iattrP->ia_flags & GPFS_IAFLAG_REPLDATA)
    bufP += sprintf(bufP, " repldata");

  /* Can't compare on the file's status bits */
  if (printCompare == 0)
  {
    if (iattrP->ia_flags & GPFS_IAFLAG_EXPOSED)
      bufP += sprintf(bufP, " exposed");
    if (iattrP->ia_flags & GPFS_IAFLAG_ILLREPLICATED)
      bufP += sprintf(bufP, " illReplicated");
    if (iattrP->ia_flags & GPFS_IAFLAG_UNBALANCED)
      bufP += sprintf(bufP, " unbalanced");
    if (iattrP->ia_flags & GPFS_IAFLAG_DATAUPDATEMISS)
      bufP += sprintf(bufP, " dataUpdateMiss");
    if (iattrP->ia_flags & GPFS_IAFLAG_METAUPDATEMISS)
      bufP += sprintf(bufP, " metaUpdateMiss");
    if (iattrP->ia_flags & GPFS_IAFLAG_ILLPLACED)
      bufP += sprintf(bufP, " illPlaced");
  }

  if (iattrP->ia_flags & GPFS_IAFLAG_COMANAGED)
    bufP += sprintf(bufP, " coManaged");
  if ((iattrP->ia_flags & GPFS_IAFLAG_READMANAGED) ||
      (iattrP->ia_flags & GPFS_IAFLAG_WRITEMANAGED) ||
      (iattrP->ia_flags & GPFS_IAFLAG_TRUNCMANAGED))
    bufP += sprintf(bufP, "-");
  if (iattrP->ia_flags & GPFS_IAFLAG_READMANAGED)
    bufP += sprintf(bufP, "R");
  if (iattrP->ia_flags & GPFS_IAFLAG_WRITEMANAGED)
    bufP += sprintf(bufP, "W");
  if (iattrP->ia_flags & GPFS_IAFLAG_TRUNCMANAGED)
    bufP += sprintf(bufP, "T");

  /* deleted files don't report a fileset id */
  if (iattrP->ia_nlink != 0)
  {
    rc = gpfs_igetfilesetname(iscanP, iattrP->ia_filesetid,
                              nbuf, sizeof(nbuf));
    if (rc != 0)
    {
      rc = errno;
      bufP += sprintf(bufP, " gpfs_igetfilesetname(%d):%s", 
                      iattrP->ia_filesetid, strerror(rc));
    }
    else if (strlen(nbuf) > 0 && strcmp(nbuf,"root") != 0)
      bufP += sprintf(bufP, " fileset=%s", nbuf);
  }

  rc = gpfs_igetstoragepool(iscanP, iattrP->ia_datapoolid,
                            nbuf, sizeof(nbuf));
  if (rc != 0)
  {
    rc = errno;
    bufP += sprintf(bufP, " gpfs_igetstoragepool(%d):%s", 
                    iattrP->ia_datapoolid, strerror(rc));
  }
  else if (strcmp(nbuf,"system") != 0)
    bufP += sprintf(bufP, " datapool='%s'", nbuf);

#ifdef GPFS_64BIT_INODES
   
  bufP += sprintf(bufP, " dev=%d,%d",
                  DevMajor(iattrP->ia_dev),
                  DevMinor(iattrP->ia_dev));

  /* Can't compare on the file's archivebits */
  if (printCompare == 0)
    if (iattrP->ia_winflags & GPFS_IWINFLAG_ARCHIVE)
      bufP += sprintf(bufP, " archive");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_HIDDEN)
    bufP += sprintf(bufP, " hidden");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_NOTINDEXED)
    bufP += sprintf(bufP, " notIndexed");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_OFFLINE)
    bufP += sprintf(bufP, " off-line");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_READONLY)
    bufP += sprintf(bufP, " read-only");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_REPARSE)
    bufP += sprintf(bufP, " reparsePoint");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_SYSTEM)
    bufP += sprintf(bufP, " system");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_TEMPORARY)
    bufP += sprintf(bufP, " temporary");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_COMPRESSED)
    bufP += sprintf(bufP, " compressed");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_ENCRYPTED)
    bufP += sprintf(bufP, " encrypted");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_SPARSE)
    bufP += sprintf(bufP, " sparse");
  if (iattrP->ia_winflags & GPFS_IWINFLAG_HASSTREAMS)
    bufP += sprintf(bufP, " hasStreams");

  /* Convert inode fields to common format */
  crsec = iattrP->ia_createtime.tv_sec;

  bufP += sprintf(bufP, " crtime %9lld.%09d",
                  crsec, iattrP->ia_createtime.tv_nsec);

  if (iattrP->ia_flags & GPFS_IAFLAG_IMMUTABLE)
    bufP += sprintf(bufP, " immutable");
  if (iattrP->ia_flags & GPFS_IAFLAG_INDEFRETENT)
    bufP += sprintf(bufP, " indefRetent");
  if (iattrP->ia_flags & GPFS_IAFLAG_SECUREDELETE)
    bufP += sprintf(bufP, " shred");
  if (iattrP->ia_flags & GPFS_IAFLAG_APPENDONLY)
    bufP += sprintf(bufP, " appendOnly");

#endif  

  printf("%s", buf);
}

static void print_xattrs(gpfs_iscan_t *iscanP,
                         const char *xattrP, unsigned int xattrLen,
                         int printCompare)
 {
   int rc;
   int i;
   const char *nameP;
   const char *valueP;
   unsigned int valueLen;
   const char *xattrBufP = xattrP;
   unsigned int xattrBufLen = xattrLen;
   int printable;

   while ((xattrBufP != NULL) && (xattrBufLen > 0))
   {
     rc = gpfs_next_xattr(iscanP, &xattrBufP, &xattrBufLen,
                          &nameP, &valueLen, &valueP);
     if (rc != 0)
     {
       rc = errno;
       fprintf(stderr, "gpfs_next_xattr: %s\n", strerror(rc));
       goto exit;
     }
     if (nameP == NULL)
       break;

     /* Eliminate gpfs.dmapi attributes for comparision
        Internal dmapi attributes are filtered from snapshots */
     if (printCompare > 1)
     {
       if (strncmp(nameP, "gpfs.dmapi.", sizeof("gpfs.dmapi.")-1) == 0)
         continue;
     }
     
     printf(" \"%s\"=", nameP);

     if (valueLen > 0)
     {
       printable = 0;
       if (valueLen > 1)
       {
         printable = 1;
         for (i = 0; i < (valueLen-1); i++)
           if (!isprint(valueP[i]))
             printable = 0;
         if (printable)
         {
           if (valueP[valueLen-1] == '\0')
             valueLen -= 1;
           else if (!isprint(valueP[valueLen-1]))
             printable = 0;
         }
       }

       printf("%s", (printable) ? "\"" : "'");
       for (i = 0; i < valueLen; i++)
       {
         if (printable)
           printf("%c", valueP[i]);
         else
           printf("%c%c", 
                  PRINTHEX((valueP[i] & 0xF0) >> 4),
                  PRINTHEX(valueP[i] & 0x0F));
       }
       printf("%s", (printable) ? "\"" : "'");
     }
  }
 exit:;
}

static int read_inodes(const char *fnameP,
                       int fset, 
                       const char *fsNameP, 
                       const char *fsetNameP, 
                       const char *snapNameP,
                       const char *prevSnapNameP,
                       int nXAttrs, const char *xattrNames[], 
                       gpfs_ino_t startinode, gpfs_ino_t endinode,
                       int printCompare)
{
  int rc = 0;
  const gpfs_iattr_t *iattrP;
  gpfs_iscan_t *iscanP = NULL;
  gpfs_fssnap_handle_t *fsP = NULL;
  gpfs_fssnap_handle_t *prevFsP = NULL;
  const gpfs_fssnap_id_t* prevFsSnapIdP = NULL;
  gpfs_fssnap_id_t prevFsSnapId;
  gpfs_snapid_t snapId;
  const char *xattrP;
  unsigned int xattrLen;

  if (fsNameP != NULL)
  {
    if (fset) 
    {
      fsP = gpfs_get_fset_snaphandle_by_name(fsNameP, fsetNameP, snapNameP);
      if (fsP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fset_snaphandle_by_name(%s,%s,%s): %s\n", 
                fsNameP, fsetNameP, snapNameP, strerror(rc));
        goto exit;
      }
    }
    else
    {
      fsP = gpfs_get_fssnaphandle_by_name(fsNameP, snapNameP);
      if (fsP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fssnaphandle_by_name(%s,%s): %s\n", 
                fsNameP, snapNameP, strerror(rc));
        goto exit;
      }
    }
  }
  else
  {
    if (fset)
    {
      fsP = gpfs_get_fset_snaphandle_by_path(fnameP);
      if (fsP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fset_snaphandle_by_path(%s): %s\n", 
                fnameP, strerror(rc));
        goto exit;
      }
    }
    else
    {
      fsP = gpfs_get_fssnaphandle_by_path(fnameP);
      if (fsP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fssnaphandle_by_path(%s): %s\n", 
                fnameP, strerror(rc));
        goto exit;
      }
    }
  }
  snapId = gpfs_get_snapid_from_fssnaphandle(fsP);
  if (snapId < 0)
  {
    rc = errno;
    fprintf(stderr, "gpfs_get_snapid_from_fssnaphandle: %s\n", strerror(rc));
    goto exit;
  }

  if (prevSnapNameP != NULL)
  {
    if (fsNameP == NULL)
    {
      fsNameP = gpfs_get_fsname_from_fssnaphandle(fsP);
      if (fsNameP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fsname_from_fssnaphandle: %s\n", strerror(rc));
        goto exit;
      }
    }
    if (fset)
    {
      if ((fsetNameP == NULL) &&
          (fnameP != NULL))
      {
        fprintf(stderr, "Must specify fileset name to use previous snapshot for scan.\n");
        rc = -1;
        goto exit;
      }
      prevFsP = gpfs_get_fset_snaphandle_by_name(fsNameP, fsetNameP, prevSnapNameP);
      if (prevFsP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fset_snaphandle_by_name(%s,%s,%s): %s\n", 
                fsNameP, fsetNameP, prevSnapNameP, strerror(rc));
        goto exit;
      }
    }
    else
    {
      prevFsP = gpfs_get_fssnaphandle_by_name(fsNameP, prevSnapNameP);
      if (prevFsP == NULL)
      {
        rc = errno;
        fprintf(stderr, "gpfs_get_fssnaphandle_by_name(%s,%s): %s\n", 
                fsNameP, prevSnapNameP, strerror(rc));
        goto exit;
      }
    }

    rc = gpfs_get_fssnapid_from_fssnaphandle(prevFsP, &prevFsSnapId);
    if (rc != 0)
    {
      rc = errno;
      fprintf(stderr, "gpfs_get_fssnapid_from_fssnaphandle: %s\n", strerror(rc));
      goto exit;
    }
    prevFsSnapIdP = &prevFsSnapId;
  }


  if (nXAttrs != 0)
    iscanP = gpfs_open_inodescan_with_xattrs(fsP, prevFsSnapIdP,
                                             nXAttrs, xattrNames, NULL);
  else
    iscanP = gpfs_open_inodescan(fsP, prevFsSnapIdP, NULL);
  if (iscanP == NULL)
  {
    rc = errno;
    fprintf(stderr, "gpfs_open_inodescan: %s\n", strerror(rc));
    goto exit;
  }

  if (startinode > 0)
  {
    rc = gpfs_seek_inode(iscanP, startinode);
    if (rc != 0)
    {
      rc = errno;
      fprintf(stderr, "gpfs_seek_inode: %s\n", strerror(rc));
      goto exit;
    }
  }

  while (1)
  {
    if (nXAttrs != 0)
      rc = gpfs_next_inode_with_xattrs(iscanP, endinode, &iattrP,
                                       &xattrP, &xattrLen);
    else
      rc = gpfs_next_inode(iscanP, endinode, &iattrP);
    if (rc != 0)
    {
      rc = errno;
      fprintf(stderr, "gpfs_next_inode: %s\n", strerror(rc));
      goto exit;
    }
    if ((iattrP == NULL) ||
        (iattrP->ia_inode > endinode))
      break;

    print_vattr(iscanP, snapId, iattrP, printCompare);

    if ((nXAttrs != 0) &&
        (xattrLen > 0))
      print_xattrs(iscanP, xattrP, xattrLen, printCompare);
    
    printf("\n");
  }
exit:
  if (iscanP)
    gpfs_close_inodescan(iscanP);
  if (prevFsP)
    gpfs_free_fssnaphandle(prevFsP);
  if (fsP)
    gpfs_free_fssnaphandle(fsP);

  return rc;
}

/* main */
int main(int argc, char *argv[])
{
  int rc;
  int i, j;
  gpfs_ino_t startinode = 0;
  gpfs_ino_t endinode = INODENUM_MAX;
  int setStart = 0;
  int setEnd = 0;
  int nXAttrs = 0;
  const char *xattrNames[MAX_XATTRS];
  int printCompare = 0;
  const char *fsNameP = NULL;
  const char *fsetNameP = NULL;
  const char *snapNameP = NULL;
  const char *prevSnapNameP = NULL;
  int fset = 0;

  /* if no files were specified, there is nothing to do */
  if (argc < 2)
    usage(argv[0]);

  for (i = 0; i < MAX_XATTRS; i++)
    xattrNames[i] = NULL;

  for (i = 2; i < argc; i++)
  {
    if (strcmp(argv[i], "-x") == 0)
    {
      nXAttrs = 0;
      for (j = i+1; j < argc; j++)
      {
        if (isdigit(argv[j][0]))
          break;
        if (argv[j][0] == '-')
          break;
        if (nXAttrs >= MAX_XATTRS)
        {
          fprintf(stderr, "Max number of xattrs is %d\n", MAX_XATTRS);
          usage(argv[0]);
        }

        xattrNames[nXAttrs] = argv[j];
        nXAttrs += 1;
        i += 1;
      }
      if (nXAttrs == 0)
        nXAttrs = -1;           /* -1 == get all xattrs */
    }

    else if (strcmp(argv[i], "-c") == 0)
      printCompare += 1;
    
    else if (strcmp(argv[i], "--fset") == 0)
      fset = 1;
    else if (strcmp(argv[i], "--fsname") == 0)
    {
      i += 1;
      if (i >= argc)
      {
        fprintf(stderr,"Missing argument to '%s'\n", argv[i-1]);
        exit(1);
      }
      fsNameP = argv[i];
    }
    else if (strcmp(argv[i], "--fsetname") == 0)
    {
      i += 1;
      if (i >= argc)
      {
        fprintf(stderr,"Missing argument to '%s'\n", argv[i-1]);
        exit(1);
      }
      fsetNameP = argv[i];
    }
    else if (strcmp(argv[i], "--snapname") == 0)
    {
      i += 1;
      if (i >= argc)
      {
        fprintf(stderr,"Missing argument to '%s'\n", argv[i-1]);
        exit(1);
      }
      snapNameP = argv[i];
    }
    else if (strcmp(argv[i], "--prevsnapname") == 0)
    {
      i += 1;
      if (i >= argc)
      {
        fprintf(stderr,"Missing argument to '%s'\n", argv[i-1]);
        exit(1);
      }
      prevSnapNameP = argv[i];
    }
      
    else if (setStart == 0)
    {
      startinode = (gpfs_ino_t) atoi(argv[i]);
      if (startinode < 0)
        startinode = 0;
      setStart = 1;
    }
    else if (setEnd == 0)
    {
      endinode = (gpfs_ino_t) atoi(argv[i]);
      if (endinode < 0)
        endinode = INODENUM_MAX;

      setEnd = 1;
    }
    else
      usage(argv[0]);
  }

  rc = read_inodes(argv[1], 
                   fset, fsNameP, fsetNameP, snapNameP, prevSnapNameP,
                   nXAttrs, xattrNames, startinode, endinode, printCompare);
  return rc;
}
