#!/bin/ksh
# Reboot a DCS3700 DCM
# Syntax:  netappdcmreboot.sh {esmSgDevice} {dcmNumber}
#   where:
#      esmSgDevice  -  is of the form /dev/sgxx
#      dcmNumber    -  0 thru 4
#

function usage
{
  print -- "Syntax:  netappdcmreboot.sh {esmSgDevice} {dcmNumber}"
  print -- "  where:"
  print -- "         esmSgDevice  -  is of the form /dev/sgxx"
  print -- "         dcmNumber    -  0 thru 4"
  print -- " "
  print -- "Example:  netappdcmreboot.sh /dev/sg5 3"
  exit
}

if [[ $# -ne 2 ]]
then
  print -- "Wrong number of arguments."
  usage
fi 

device=$1

# Setup the dcmBitMask field:
#   DCM0 = 01h
#   DCM1 = 02h
#   DCM2 = 04h
#   DCM3 = 08h
#   DCM4 = 10h

case $2 in
  0) dcmBitMask=01 ;;  
  1) dcmBitMask=02 ;;
  2) dcmBitMask=04 ;;
  3) dcmBitMask=08 ;;
  4) dcmBitMask=10 ;;
  *) usage ;;
esac
  

#                        String  Page  Request     Magic     Parameters
#                        Out     Len               NUM          #0     #1       #2       #3         #4
#                                                            DCM#   
#                          --    ----- ----------- --------  -- ------ -------- -------- -------- --------
#/usr/bin/sg_senddiag --raw=04 00 0028  00000004    424F4F54  00000000 00000000 00000000 00000000 00000000
#
#                         Parameters
#                           #5      #6       #7     
#                         -------- -------- --------
#                         00000000 00000000 00000000

# where: CodeName            Request  Magic    Parameters
#        Encl Power Down     00000001 5057444E num of secs to delay 10-300
#        Hold DCMS in Reset  00000002 484F4C44 DCM bitmask
#        Rel DCMs from Reset 00000003 52454C45 DCM bitmask
#        Reboot DCMs         00000004 424F4F54 DCM bitmask

print -- "/usr/bin/sg_senddiag --pf --raw=04000028 00000004 424F4F54 000000${dcmBitMask} 00000000 00000000 00000000 00000000 00000000 00000000 00000000 ${device}"

/usr/bin/sg_senddiag --pf --raw="04 00 00 28 00 00 00 04 42 4F 4F 54 00 00 00 ${dcmBitMask} 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00" ${device}
exit $?
}

