/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
/* @(#)66       1.16.1.14  src/avs/fs/mmfs/ts/kernext/gpl-linux/ktrccalls.c, mmfs, avs_rhrz, rhrzs018a 3/11/14 17:18:17 */

/**************************************************************************
 *
 * Common kernel trace functions
 *
 **************************************************************************/

#ifndef GPFS_PRINTF

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifdef GPFS_LINUX
#ifndef GPL_KBUILD
/* We have to define __SMALL_STACK for s390x kernels with 8K stacks, otherwise
   THREAD_SIZE will be set wrong. Without a correct THREAD_SIZE, in_interrupt()
   will return wrong results.
   Since we use our own build environment we have to do this by ourself. The
   kernel configuration is automatically included via -include when compiling
   this source file. */
#if defined(GPFS_ARCH_S390X) && defined(CONFIG_SMALL_STACK)
#  define __SMALL_STACK
#endif
#endif /* GPL_KBUILD */

#define Kprint printk
#include <Shark-gpl.h>

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <stdarg.h>
#include <linux/string.h>

#include <Trace.h>
#include <lxtrace.h>
#include <verdep.h>
#include <cxiAtomic.h>

#ifdef LXTRACE_USE_RELAY
extern int rl_trc_write(const char *bufP, size_t nBytes, int nb);
extern int isRLEnabled();
#endif

#if LINUX_KERNEL_VERSION > 2060900
#include <linux/hardirq.h> /* interrupt_context */
#else
#include <asm/hardirq.h> /* interrupt_context */
#endif

unsigned int interrupt_context()
{
  return in_interrupt();
}

#elif defined(GPFS_SOLARIS)

#include <Trace.h>
#include <lxtrace.h>
#include <sys/cmn_err.h>
#include <sys/cpuvar.h>
#include <string.h>
#include <stdarg.h>

#define Kprint(...) cmn_err (CE_NOTE,__VA_ARGS__);
#define MIN(a,b)  ((a)<(b) ? (a) : (b))
#define memcpy(_a,_b,_c) bcopy((_b), (_a), (_c))

unsigned int interrupt_context()
{
  return CPU_ON_INTR(CPU);
}

typedef unsigned int Boolean;
#define false 0
#define true 1

#endif /* GPFS_SOLARIS */

extern void trc_append_record(const void*, size_t);

void
ktrc_append_record(char *dst, const void* src, int nBytes, int useMemcpy)
{
  /* memcpy is used for overwrite mode or for relayfs. useMemcpy is for
     pre-relayfs code. */
  if (useMemcpy)
  {
    memcpy(dst, src, nBytes);
  }
  else
    trc_append_record(src, nBytes);
}

#ifdef SHARED_TRC_BUFF
inline void ktrc_fill_header(newHdr_t *hdr, int hookword,
                             int nArgs, int pos, int len)

{
  hdr->hdr.trHook = hookword;
  hdr->hdr.trNArgs = nArgs;
  hdr->hdr.trSPos = pos;
  hdr->hdr.trSLen = len;
}
#endif

inline void ktrc_fill_dataheader(trc_datahdr_t *hdr, int hookword,
                             int nArgs, int pos, int len)

{
  hdr->trHook = hookword;
  hdr->trNArgs = nArgs;
  hdr->trSPos = pos;
  hdr->trSLen = len;
}

#if defined(__64BIT__) || defined(_LP64)
# define Int64 long long
# define ARGLEN 8
# define ARGTYPE Int64
#else
# define Int32 int
# define ARGLEN 4
# define ARGTYPE Int32
#endif

char stringPadding[8];

/* The following 4 functions are required to support pre-SHARED_TRC_BUFF code. */
extern trcdev_header_t   lxthe;
extern void trc_append_record(const void*, size_t);
extern int trc_start_record(trc_datahdr_t *, size_t, int);
extern void trc_end_record();

extern Boolean isTraced(uint);

#ifdef SHARED_TRC_BUFF
/* Get the current clock cycle count. Duplicated in tracelin.c */
static inline UInt64 __readCycleCounter()
{
#if defined(GPFS_ARCH_X86_64)
  UInt32 low, high;
  __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high));
  return (low | ((UInt64)(high) << 32));
#elif defined(GPFS_ARCH_PPC64) || defined(GPFS_ARCH_POWER)
  UInt64 val;
  __asm__ __volatile__("mftb %0" : "=r" (val));
  return val;
#elif defined(GPFS_ARCH_S390X)
  UInt64 val;
  __asm__ __volatile__("stck 0(%1)" : "=m" (val) : "a" (&val) : "cc");
  return val;
#else
  #error "Can't get CPU count for this linux architecture"
#endif
}

#define rdtscall(val) \
  ((val) = __readCycleCounter())


/* Global variables exported from the trace module
   Used by the SHARED_TRC_BUFFER code. */

char* KTraceBufferPtr = NULL;
Int64 KTraceBufferSize = -1;
Int64* KTraceOffsetPtr = NULL;
int *KTraceWrapPtr = NULL;
int KTraceMode = BLOCKING;
int *KTraceModeP = &KTraceMode;
int KTraceIsOn = 0;


/* To get useful Time of Day (TOD) traces, there must be
   a reasonable number of them. Make sure there are enough in each wrap.
   We have matching code in gpfslinux.C, ktrccalls.c and tracelin.c (for when we
   are copying traces into the shared buffer via ioctl). No synchronization is
   done to track how many TOD timestamps we are emitting, so each trace path
   will emit TODPerWrap TOD traces per wrap. */
extern int KprevWrapCount ;
#define  KTODPerWrap 4
extern int KTODThisWrap;
extern int *KTODThisWrapP;

/* Checks if it's time to emit a TOD trace; if so, emit one.
 We reset KprevWrapCount whenever we stop traces (this means that cutting
 traces will also force a reset, but I think this is a feature not a bug). */
static void checkTODTrace()
{
  struct timeval trTime;
  newHdr_t *hdrP;
  char tBuf[LXTRACE_MAX_DATA+1];
  long long sec, usec;
  int csSuccess;
  long long endStamp = LXTRACE_TOD_END_MAGIC;
  if (KprevWrapCount != *KTraceWrapPtr)
    KTODThisWrap = 0;
  /* To make sure we don't emit too many TOD traces,
     force it to use compare and swap to track the number of
     emitted TOD traces. */

  do
  {
    if (*KTraceOffsetPtr < KTODThisWrap * (KTraceBufferSize / KTODPerWrap))
      return;

    csSuccess = ATOMIC_SWAP((atomic_p)KTODThisWrapP,
                            &KTODThisWrap, KTODThisWrap + 1);
  } while (!csSuccess);

  KTODThisWrap++;
  KprevWrapCount = *KTraceWrapPtr;
  memset(tBuf, '9', LXTRACE_MAX_DATA);
  hdrP = (newHdr_t*)tBuf;
  hdrP->hdr.trHook = LXTRACE_TOD_MAGIC;
  hdrP->hdr.trNArgs = 2;
  hdrP->hdr.trSPos = _TR_FORMAT_I;
  hdrP->hdr.trSLen = 0;

  do_gettimeofday(&trTime);
  sec = (long long)trTime.tv_sec;
  usec = (long long)trTime.tv_usec;
  memcpy(tBuf+sizeof(newHdr_t), &(sec), sizeof(long long));
  memcpy(tBuf+sizeof(newHdr_t)+ sizeof(long long), &(usec), sizeof(long long));
  memcpy(tBuf+sizeof(newHdr_t)+ 2*sizeof(long long), &(endStamp), sizeof(long long));
  appendKernelTrace((newHdr_t*)tBuf, sizeof(newHdr_t) + 3*sizeof(long long));
}

/* Returns the offset in the buffer that can take copyLen bytes.
   Add that to KTraceBufferPtr to find the address to copy things into.*/
#define ROUNDUP(X,Y) ( (Y) * ( ((X)+(Y)-1) / (Y) ) )
static Int64 reserveSharedBufferSpace(int copyLen)
{
  char* traceP;
  int reserveLen;
  Int64 oldOffset = 0;
  Int64 newOffset = 0;
  Boolean csSuccess;
  Int64 bufRemaining;
  /* Round trace record size up to a cache line boundary */
  reserveLen = ROUNDUP(copyLen, CACHE_LINE);

  /* Obtain a slot in the trace buffer for the current record. This
     assignment is done outside the loop since ATOMIC_SWAP automatically
     updates oldOffset if the swap fails. */
  oldOffset = *KTraceOffsetPtr;
  do
  {
    newOffset = oldOffset + reserveLen;
    bufRemaining = KTraceBufferSize - newOffset;
    if (unlikely(bufRemaining <= 0))
      newOffset -= KTraceBufferSize;
    csSuccess = ATOMIC_SWAP64((atomic_p)KTraceOffsetPtr,
                            &oldOffset, newOffset);
  } while (!csSuccess);
  return oldOffset;
}

inline void fillNewHdr(void* ptrP, int trRecLen)
{
  typedef struct
  {
    newtrc_header_t h0;
    trc_datahdr_t hdr;
  } newHdr_t;
  newHdr_t *newHdrP = (newHdr_t *) ptrP;

  newHdrP->h0.trMagic = LXTRACE_SWAP_MAGIC;
  rdtscall(newHdrP->h0.trTime);
  newHdrP->h0.trCPU = smp_processor_id();
  newHdrP->h0.trProcess = (pid_t) current->pid;
  newHdrP->h0.trLength = trRecLen;
  newHdrP->h0.trWrapPoint = *KTraceWrapPtr;
}

/* Copy copyLen bytes from newHdrP to KTraceBufferPtr
   starting at oldOffset. */
void copyToSharedBuffer(newHdr_t* newHdrP, int copyLen, Int64 oldOffset)
{
  /* Copy the trace record into the slot just obtained.  It may wrap
     around the trace buffer. */
  char* traceP;
  Int64 bufRemaining;
  traceP = KTraceBufferPtr + oldOffset;;
  bufRemaining = KTraceBufferSize - oldOffset;

  if (copyLen <= bufRemaining)
  {
    memcpy(traceP, newHdrP, copyLen);
  }
  else
  {
    memcpy(traceP, newHdrP, bufRemaining);
    traceP = KTraceBufferPtr;
    memcpy(traceP, (char*)newHdrP + bufRemaining, copyLen-bufRemaining);
  }

  /* If we are wrapping or made exactly to the end, update the wrap counter so the next entry
     will know it has wrapped.   
     Note that some threads that came after us may already have created trace record using the 
     old counter before we got to increment or vise-versa, some threads that came before us but 
     got heldup uptil now would be using the incremented counter. This is acceptable risk as 
     correcting that would require atomicity in setting TraceOffset and wrap counter. Lxtrace will 
     identify records in such category. */
  if (copyLen >= bufRemaining)
    *KTraceWrapPtr = *KTraceWrapPtr + 1;
}

/* Copy trace record into trace buffer. Duplicated in tracelin.c */
static void appendKernelTrace(newHdr_t* newHdrP, int trRecLen)
{
  char* traceP;
  int copyLen;
  Int64 oldOffset;
  /* Round trace record size up to a cache line boundary */
  copyLen = trRecLen;

  oldOffset = reserveSharedBufferSpace(copyLen);

  /* Fill in the timestamp and other header fields */
  fillNewHdr(newHdrP, trRecLen);

  copyToSharedBuffer(newHdrP, copyLen, oldOffset);

}
#endif /* SHARED_TRC_BUFF */

static void
_STraceArgs(int* trRecLenP, int* stringLenP, int nArgs,
            int pos, va_list listP, char *tBuf)
{
  int dataLen = 0;
  int i;
  ARGTYPE tmpint;
  char *s;
  int stringLen;
  int stringPadLen;
  int useMemcpy = 0;  /* Replaces the old "relay" variable. */
  int hdrSize;
#ifdef SHARED_TRC_BUFF
  useMemcpy = 1;
  if (*KTraceModeP == OVERWRITE)
  {
    dataLen = sizeof(newHdr_t);
    hdrSize = sizeof(newHdr_t);
  }
  else
#endif /* SHARED_TRC_BUFF */
  {
    dataLen = sizeof(trc_datahdr_t);
    hdrSize = sizeof(trc_datahdr_t);
    if(tBuf != NULL)
    {
      useMemcpy = 1;
    }
  }

  /* Handle argument lists that include a string parameter */
  if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
  {
    /* Items (if any) preceeding the string argument */
    for (i = 0; i < pos; i++)
    {
      tmpint = va_arg(listP, ARGTYPE);
      ktrc_append_record(tBuf+dataLen, &tmpint, ARGLEN, useMemcpy);
      dataLen += ARGLEN;
    }
    /* Copy the string, making sure it does not overflow the buffer */
    s = va_arg(listP, char*);
    if (s == NULL)
      s = "(NULL)";
    if (s < (char*)4096) /* bad address */
    {
      #ifdef SHARED_TRC_BUFF
      if (*KTraceModeP == OVERWRITE)
        Kprint("_STrace: bad address %p hook 0x%X\n", s, ((trc_datahdr_t*)tBuf)->trHook);
      else
      #else
      Kprint("_STrace: bad address %p hook 0x%X\n", s, lxthe.hdrP->trHook);
      #endif /* SHARED_TRC_BUF */
      s = "<bad address>";
    }
    stringLen = strlen(s);
    stringLen = MIN(stringLen,
                    LXTRACE_MAX_DATA - hdrSize -
                      (nArgs*ARGLEN) - 1 - (ARGLEN-1));
    ktrc_append_record(tBuf+dataLen, s, stringLen, useMemcpy);
    dataLen += stringLen;
    stringPadLen = ARGLEN - (stringLen%ARGLEN);
    ktrc_append_record(tBuf+dataLen, stringPadding, stringPadLen, useMemcpy);
    *stringLenP = stringLen + stringPadLen;
    dataLen += stringPadLen;
    /* Append items following string argument */
    for (i = pos; i < nArgs; i++)
    {
      tmpint = va_arg(listP, ARGTYPE);
      ktrc_append_record(tBuf+dataLen, &tmpint, ARGLEN, useMemcpy);
      dataLen += ARGLEN;
    }
  }
  else /* !IS_SFORMAT */
  {
    /* Place the fixed parameters in the temporary trace buffer */
    for (i = 0; i < nArgs; i++)
    {
      tmpint = va_arg(listP, ARGTYPE);
      ktrc_append_record(tBuf+dataLen, &tmpint, ARGLEN, useMemcpy);
      dataLen += ARGLEN;
    }
    *stringLenP = 0;
  }
  /* Append the float argument */
  if (pos == _TR_FORMAT_F)
  {
    /* Although the argument is really a double, don't tell the compiler,
       so that it will not generate code using floating point hardware
       that is not supposed to be used in the kernel. */
    /* double tmpdbl = va_arg(listP, double); */
    unsigned long long tmpdbl = va_arg(listP, unsigned long long);
    ktrc_append_record(tBuf+dataLen, &tmpdbl, sizeof(tmpdbl), useMemcpy);
    dataLen += sizeof(tmpdbl);
  }
  *trRecLenP = dataLen;
  /* DBGASSERT(*trRecLenP <= LXTRACE_MAX_DATA); */
}

static void
_DTraceArgs(int* trRecLenP, int* stringLenP, int nArgs,
            int pos, va_list listP, char *tBuf)
{
  /* Only integer arguments */
  int dataLen = 0;
  int i;
  ARGTYPE tmpint;
  char *s;
  int stringLen;
  int stringPadLen;
  int useMemcpy = 0;  /* Replaces the old "relay" variable. */
  int hdrSize;
#ifdef SHARED_TRC_BUFF
  useMemcpy = 1;
  if (*KTraceModeP == OVERWRITE)
  {
    dataLen = sizeof(newHdr_t);
    hdrSize = sizeof(newHdr_t);
  }
  else
#endif /* SHARED_TRC_BUFF */
  {
    dataLen = sizeof(trc_datahdr_t);
    hdrSize = sizeof(trc_datahdr_t);
    if(tBuf != NULL)
    {
      useMemcpy = 1;
    }
  }
  /* Place the fixed parameters in the temporary trace buffer */
  for (i = 0; i < nArgs; i++)
  {
    tmpint = va_arg(listP, ARGTYPE);
    ktrc_append_record(tBuf+dataLen, &tmpint, ARGLEN, useMemcpy);
    dataLen += ARGLEN;
  }
  *stringLenP = 0;

  *trRecLenP = dataLen;
  /* DBGASSERT(*trRecLenP <= LXTRACE_MAX_DATA); */
}

void _STraceNB(int hookword, int nArgs, int pos, ...)
{
  int recLen;
  int rc;
  va_list listP;
  int trRecLen = 0;
  int stringLen;
  char tBuf[LXTRACE_MAX_DATA+1];
#ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
#endif /* SHARED_TRC_BUFF */
  /* These variables need to be defined in case we are in blocking mode */
  trc_datahdr_t hdr;
  trc_datahdr_t *hdrP;

  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Test for trace formats that aren't supported yet */
  if ((pos == _TR_FORMAT_I) && (nArgs > LXTRACE_MAX_FORMAT_SUBS))
  {
#ifdef DBGASSERTS
    Kprint("_STrace: too many arguments (hook %X)\n", hookword);
#endif /* DBGASSERTS */
    return;
  }
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    nHdrP = (newHdr_t*)tBuf;
    ktrc_fill_header(nHdrP, hookword, nArgs, pos, 0);
    va_start(listP, pos);
    _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
    va_end(listP);
    nHdrP->hdr.trSLen = stringLen;
    checkTODTrace();
    appendKernelTrace((newHdr_t*)tBuf, trRecLen);
  }
  else
#endif /* SHARED_TRC_BUFF */
  {
#ifdef LXTRACE_USE_RELAY
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      va_start(listP, pos);
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
      va_end(listP);
      hdrP->trSLen = stringLen;  
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      va_start(listP, pos);
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, NULL);
      va_end(listP);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen;
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _STrace(int hookword, int nArgs, int pos, ...)
{
  int recLen;
  int rc;
  va_list listP;
  int trRecLen = 0;
  int stringLen;
  char tBuf[LXTRACE_MAX_DATA+1];
  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;

  /* Test for trace formats that aren't supported yet */
  if ((pos == _TR_FORMAT_I) && (nArgs > LXTRACE_MAX_FORMAT_SUBS))
  {
#ifdef DBGASSERTS
    Kprint("_STrace: too many arguments (hook %X)\n", hookword);
#endif /* DBGASSERTS */
    return;
  }
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    nHdrP = (newHdr_t*)tBuf;
    ktrc_fill_header(nHdrP, hookword, nArgs, pos, 0);
    va_start(listP, pos);
    _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
    va_end(listP);
    nHdrP->hdr.trSLen = stringLen;
    checkTODTrace();
    appendKernelTrace((newHdr_t*)tBuf, trRecLen);
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      va_start(listP, pos);
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
      va_end(listP);
      hdrP->trSLen = stringLen;  
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      va_start(listP, pos);
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, NULL);
      va_end(listP);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTraceNB(int hookword, int nArgs, int pos, ...)
{
  int recLen;
  int rc;
  va_list listP;
  int trRecLen = 0;
  int stringLen;
  char tBuf[LXTRACE_MAX_DATA+1];
#ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
#endif /* SHARED_TRC_BUFF */
  /* These variables need to be defined in case we are in blocking mode */
  trc_datahdr_t hdr;
  trc_datahdr_t *hdrP;

  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Test for trace formats that aren't supported yet */
  if ((pos == _TR_FORMAT_I) && (nArgs > LXTRACE_MAX_FORMAT_SUBS))
  {
#ifdef DBGASSERTS
    Kprint("_STrace: too many arguments (hook %X)\n", hookword);
#endif /* DBGASSERTS */
    return;
  }
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    nHdrP = (newHdr_t*)tBuf;
    ktrc_fill_header(nHdrP, hookword, nArgs, pos, 0);
    va_start(listP, pos);
    _DTraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
    va_end(listP);
    nHdrP->hdr.trSLen = stringLen;
    checkTODTrace();
    appendKernelTrace((newHdr_t*)tBuf, trRecLen);
  }
  else
#endif /* SHARED_TRC_BUFF */
  {
#ifdef LXTRACE_USE_RELAY
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      va_start(listP, pos);
      _DTraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
      va_end(listP);
      hdrP->trSLen = stringLen;  
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      va_start(listP, pos);
      _DTraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, NULL);
      va_end(listP);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen;
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace(int hookword, int nArgs, int pos, ...)
{
  int recLen;
  int rc;
  va_list listP;
  int trRecLen = 0;
  int stringLen;
  char tBuf[LXTRACE_MAX_DATA+1];
#ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
#endif /* SHARED_TRC_BUFF */
  /* These variables need to be defined in case we are in blocking mode */
  trc_datahdr_t hdr;
  trc_datahdr_t *hdrP;

  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Test for trace formats that aren't supported yet */
  if ((pos == _TR_FORMAT_I) && (nArgs > LXTRACE_MAX_FORMAT_SUBS))
  {
#ifdef DBGASSERTS
    Kprint("_STrace: too many arguments (hook %X)\n", hookword);
#endif /* DBGASSERTS */
    return;
  }
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    nHdrP = (newHdr_t*)tBuf;
    ktrc_fill_header(nHdrP, hookword, nArgs, pos, 0);
    va_start(listP, pos);
    _DTraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
    va_end(listP);
    nHdrP->hdr.trSLen = stringLen;
    checkTODTrace();
    appendKernelTrace((newHdr_t*)tBuf, trRecLen);
  }
  else
#endif /* SHARED_TRC_BUFF */
  {
#ifdef LXTRACE_USE_RELAY
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      va_start(listP, pos);
      _DTraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, tBuf);
      va_end(listP);
      hdrP->trSLen = stringLen;  
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      va_start(listP, pos);
      _DTraceArgs(&trRecLen, &stringLen, nArgs, pos, listP, NULL);
      va_end(listP);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen;
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace0(int hookword)
{
  int recLen;
  int rc;
  int trRecLen = 0;
  int stringLen = 0;

  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;
  
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    Int64 offset;
    trRecLen = sizeof(newHdr_t) + 0*sizeof(Int64);
    checkTODTrace();
    offset = reserveSharedBufferSpace(trRecLen);
    if(likely(KTraceBufferSize - offset >= trRecLen))
    {
      /* The buffer doesn't wrap */
      nHdrP = (newHdr_t*)(KTraceBufferPtr + offset);
      ktrc_fill_header(nHdrP, hookword, 0, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
    }
    else
    {
      /* Buffer wraps during this record, so use a staging buffer */
      char tBuf[LXTRACE_MAX_DATA+1];
      nHdrP = (newHdr_t*)tBuf;
      ktrc_fill_header(nHdrP, hookword, 0, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      copyToSharedBuffer(nHdrP, trRecLen, offset);
    }
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    char tBuf[LXTRACE_MAX_DATA+1];
    int pos;
    int nArgs;
    pos = _TR_FORMAT_I;
    nArgs = 0;
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      trRecLen = sizeof(trc_datahdr_t) + 0*sizeof(Int64);
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, NULL, NULL);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace1(int hookword, Int64 a1)
{
  int recLen;
  int rc;
  int trRecLen = 0;
  int stringLen = 0;
  const int nArgs = 1;
  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  Int64 *buffP;
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;
  
  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;
  
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    Int64 offset;
    trRecLen = sizeof(newHdr_t) + 1*sizeof(Int64);
    checkTODTrace();
    offset = reserveSharedBufferSpace(trRecLen);
    if(likely(KTraceBufferSize - offset >= trRecLen))
    {
      /* The buffer doesn't wrap */
      nHdrP = (newHdr_t*)(KTraceBufferPtr + offset);
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      //      memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(Int64));
    }
    else
    {
      /* Buffer wraps during this record, so use a staging buffer */
      char tBuf[LXTRACE_MAX_DATA+1];
      nHdrP = (newHdr_t*)tBuf;
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      //      memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(Int64));
      copyToSharedBuffer(nHdrP, trRecLen, offset);
    }
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    char tBuf[LXTRACE_MAX_DATA+1];
    int pos;
    pos = _TR_FORMAT_I;
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      buffP = (Int64*)((char *)(hdrP) + sizeof(trc_datahdr_t));
      buffP[0] = a1;
      trRecLen = sizeof(trc_datahdr_t) + 1*sizeof(Int64);
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, NULL, NULL);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace2(int hookword, Int64 a1, Int64 a2)
{
  int recLen;
  int rc;
  int trRecLen = 0;
  int stringLen = 0;
  const int nArgs = 2;

  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  Int64 *buffP;
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;
  
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    Int64 offset;
    trRecLen = sizeof(newHdr_t) + 2*sizeof(Int64);
    checkTODTrace();
    offset = reserveSharedBufferSpace(trRecLen);
    if(likely(KTraceBufferSize - offset >= trRecLen))
    {
      /* The buffer doesn't wrap */
      nHdrP = (newHdr_t*)(KTraceBufferPtr + offset);
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(Int64)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
    }
    else
    {
      /* Buffer wraps during this record, so use a staging buffer */
      char tBuf[LXTRACE_MAX_DATA+1];
      nHdrP = (newHdr_t*)tBuf;
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
      copyToSharedBuffer(nHdrP, trRecLen, offset);
    }
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    char tBuf[LXTRACE_MAX_DATA+1];
    int pos;
    pos = _TR_FORMAT_I;
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      buffP = (Int64*)((char *)(hdrP) + sizeof(trc_datahdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      trRecLen = sizeof(trc_datahdr_t) + 2*sizeof(Int64);
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, NULL, NULL);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace3(int hookword, Int64 a1, Int64 a2, Int64 a3)
{
  int recLen;
  int rc;
  int trRecLen = 0;
  int stringLen = 0;
  const int nArgs = 3;

  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  Int64 *buffP;
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;
  
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    Int64 offset;
    trRecLen = sizeof(newHdr_t) + 3*sizeof(Int64);
    checkTODTrace();
    offset = reserveSharedBufferSpace(trRecLen);
    if(likely(KTraceBufferSize - offset >= trRecLen))
    {
      /* The buffer doesn't wrap */
      nHdrP = (newHdr_t*)(KTraceBufferPtr + offset);
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(Int64)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 2*sizeof(Int64)), &a3, sizeof(int)); */
    }
    else
    {
      /* Buffer wraps during this record, so use a staging buffer */
      char tBuf[LXTRACE_MAX_DATA+1];
      nHdrP = (newHdr_t*)tBuf;
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 2*sizeof(Int64)), &a3, sizeof(int)); */
      copyToSharedBuffer(nHdrP, trRecLen, offset);
    }
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    char tBuf[LXTRACE_MAX_DATA+1];
    int pos;
    pos = _TR_FORMAT_I;
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      buffP = (Int64*)((char *)(hdrP) + sizeof(trc_datahdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      trRecLen = sizeof(trc_datahdr_t) + 3*sizeof(Int64);
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, NULL, NULL);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace4(int hookword, Int64 a1, Int64 a2, Int64 a3, Int64 a4)
{
  int recLen;
  int rc;
  int trRecLen = 0;
  int stringLen = 0;
  const int nArgs = 4;

  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  Int64 *buffP;
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;
  
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    Int64 offset;
    trRecLen = sizeof(newHdr_t) + 4*sizeof(Int64);
    checkTODTrace();
    offset = reserveSharedBufferSpace(trRecLen);
    if(likely(KTraceBufferSize - offset >= trRecLen))
    {
      /* The buffer doesn't wrap */
      nHdrP = (newHdr_t*)(KTraceBufferPtr + offset);
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      buffP[3] = a4;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(Int64)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 2*sizeof(Int64)), &a3, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 3*sizeof(Int64)), &a4, sizeof(int)); */
    }
    else
    {
      /* Buffer wraps during this record, so use a staging buffer */
      char tBuf[LXTRACE_MAX_DATA+1];
      nHdrP = (newHdr_t*)tBuf;
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      buffP[3] = a4;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
      copyToSharedBuffer(nHdrP, trRecLen, offset);
    }
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    char tBuf[LXTRACE_MAX_DATA+1];
    int pos;
    pos = _TR_FORMAT_I;
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      buffP = (Int64*)((char *)(hdrP) + sizeof(trc_datahdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      buffP[3] = a4;
      trRecLen = sizeof(trc_datahdr_t) + 4*sizeof(Int64);
      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, NULL, NULL);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}

void _DTrace5(int hookword, Int64 a1, Int64 a2, Int64 a3, Int64 a4, Int64 a5)
{
  int recLen;
  int rc;
  int trRecLen = 0;
  int stringLen = 0;
  const int nArgs = 5;

  #ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
  #endif
  Int64 *buffP;
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;
  
#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    Int64 offset;
    trRecLen = sizeof(newHdr_t) + 5*sizeof(Int64);
    checkTODTrace();
    offset = reserveSharedBufferSpace(trRecLen);
    if(likely(KTraceBufferSize - offset >= trRecLen))
    {
      /* The buffer doesn't wrap */
      nHdrP = (newHdr_t*)(KTraceBufferPtr + offset);
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      buffP[3] = a4;
      buffP[4] = a5;
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t)), &a1, sizeof(Int64)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + sizeof(Int64)), &a2, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 2*sizeof(Int64)), &a3, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 3*sizeof(Int64)), &a4, sizeof(int)); */
      /* memcpy(((char *)(nHdrP) + sizeof(newHdr_t) + 4*sizeof(Int64)), &a5, sizeof(int)); */
    }
    else
    {
      /* Buffer wraps during this record, so use a staging buffer */
      char tBuf[LXTRACE_MAX_DATA+1];
      nHdrP = (newHdr_t*)tBuf;
      ktrc_fill_header(nHdrP, hookword, nArgs, _TR_FORMAT_I, 0);
      fillNewHdr(nHdrP, trRecLen);
      buffP = (Int64*)((char *)(nHdrP) + sizeof(newHdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      buffP[3] = a4;
      buffP[4] = a5;
      copyToSharedBuffer(nHdrP, trRecLen, offset);
    }
  }
  else
#endif /* SHARED_TRC_BUF */ 
  {
#ifdef LXTRACE_USE_RELAY
    char tBuf[LXTRACE_MAX_DATA+1];
    int pos;
    pos = _TR_FORMAT_I;
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, nArgs, pos, 0);
      buffP = (Int64*)((char *)(hdrP) + sizeof(trc_datahdr_t));
      buffP[0] = a1;
      buffP[1] = a2;
      buffP[2] = a3;
      buffP[3] = a4;
      buffP[4] = a5;
      trRecLen = sizeof(trc_datahdr_t) + 5*sizeof(Int64);

      rl_trc_write(tBuf, trRecLen, 1);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    ktrc_fill_dataheader(&hdr, hookword, nArgs, pos, 0);
    if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
      recLen = LXTRACE_MAX_DATA;
    else
    {
      recLen = sizeof(hdr) + nArgs*ARGLEN;
      if (pos == _TR_FORMAT_F)
        recLen += ARGLEN;
    }
    rc = trc_start_record(&hdr, recLen, true);
    /* If the header was successfully written, collect arguments directly into
       the trace buffer */
    if (rc == 0)
    {
      _STraceArgs(&trRecLen, &stringLen, nArgs, pos, NULL, NULL);
      /* Patch the string and record lengths now that the string has been
         copied */
      lxthe.hdrP->trSLen = stringLen;
      lxthe.tHdrP->trLength = trRecLen; 
      /* Trace record complete */
      trc_end_record();
    }
  }
}


void _XTraceNB(int hookword, const char *fmt, ...)
{
  int rc;
  va_list vargs;
  int stringLen = 0;
  char tBuf[LXTRACE_MAX_DATA+1];
  int trLength = 0;
#ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
#endif
  trc_datahdr_t hdr;
  trc_datahdr_t *hdrP;

  //we need to use current->pid but that's not available in interrupt context
  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;

#ifdef SHARED_TRC_BUFF
  if (*KTraceModeP == OVERWRITE)
  {
    nHdrP = (newHdr_t*)tBuf;
    ktrc_fill_header(nHdrP, hookword, 0, _TR_FORMAT_X, -1);
    va_start(vargs, fmt);
    stringLen = vsnprintf(tBuf+sizeof(newHdr_t),
                          LXTRACE_MAX_DATA-sizeof(newHdr_t), fmt, vargs) + 1;
    va_end(vargs);

    if (stringLen > LXTRACE_MAX_DATA-sizeof(newHdr_t))
    {
      Kprint("_XTraceNB: argument too long.  len=%d max=%d hook=0x%X\n",
             stringLen, (int)(LXTRACE_MAX_DATA-sizeof(newHdr_t)-1), hookword);
      stringLen = LXTRACE_MAX_DATA-sizeof(newHdr_t);
    }
    
    /* Patch the string and record lengths now that vsnprintf has calculated
     the length that it formatted */
    nHdrP->hdr.trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
    trLength = sizeof(*nHdrP) + nHdrP->hdr.trSLen;
    checkTODTrace();
    appendKernelTrace((newHdr_t*)tBuf, trLength);
  }
  else
#endif /* SHARED_TRC_BUF */
  {
#ifdef LXTRACE_USE_RELAY
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, 0, _TR_FORMAT_X, -1);
    
      va_start(vargs, fmt);
      stringLen = vsnprintf(tBuf+sizeof(trc_datahdr_t),
                            LXTRACE_MAX_DATA-sizeof(trc_datahdr_t), fmt, vargs) + 1;
      va_end(vargs);
    
      if (stringLen > LXTRACE_MAX_DATA-sizeof(trc_datahdr_t))
      {
        Kprint("_XTraceNB: argument too long.  len=%d max=%d hook=0x%X\n",
               stringLen, (int)(LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)-1), hookword);
        stringLen = LXTRACE_MAX_DATA-sizeof(trc_datahdr_t);
      }
      /* Patch the string and record lengths now that vsnprintf has calculated
         the length that it formatted */
      hdrP->trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
      trLength = sizeof(hdr) + hdrP->trSLen;
      rl_trc_write(tBuf, trLength, 1);
      return;
    }
#endif
  
    /* Build a data header and append it to the trace file.  Since the length
       is not yet known, use the maximum.  It will be patched to the correct
       value later. */
    ktrc_fill_dataheader(&hdr, hookword, 0, _TR_FORMAT_X, -1);
    rc = trc_start_record(&hdr, LXTRACE_MAX_DATA, true);
  
    /* If the header was successfully written, format the string directly
       into the trace buffer */
    if (rc == 0)
    {
      va_start(vargs, fmt);
      stringLen = vsnprintf(lxthe.writeBuf.nextP,
                            LXTRACE_MAX_DATA-sizeof(trc_datahdr_t), fmt, vargs) + 1;
      va_end(vargs);
      if (stringLen > LXTRACE_MAX_DATA-sizeof(trc_datahdr_t))
      {
        Kprint("_XTraceNB: argument too long.  len=%d max=%d hook=0x%X\n",
               stringLen, (int)(LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)-1), hookword);
        stringLen = LXTRACE_MAX_DATA-sizeof(trc_datahdr_t);
      }
      /* Patch the string and record lengths now that vsnprintf has calculated
         the length that it formatted */
      lxthe.hdrP->trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
      lxthe.tHdrP->trLength = sizeof(hdr) + lxthe.hdrP->trSLen;
      /* Advance pointer into trace buffer by the length of the string just
         appended */
      lxthe.writeBuf.nextP += lxthe.hdrP->trSLen;
      /* Trace record complete */
      trc_end_record();

    }
  }
}

void _XTrace(int hookword, const char *fmt, ...)
{
  int rc;
  va_list vargs;
  int stringLen = 0;
  char tBuf[LXTRACE_MAX_DATA+1];
  int trLength = 0;
#ifdef SHARED_TRC_BUFF
  newHdr_t *nHdrP;
#endif
  trc_datahdr_t *hdrP;
  trc_datahdr_t hdr;

  //we need to use current->pid but that's not available in interrupt context
  if (interrupt_context())
    return;
  if (!isTraced(hookword))
    return;

#ifdef SHARED_TRC_BUFF
  if(*KTraceModeP == OVERWRITE)
  {

    nHdrP = (newHdr_t*)tBuf;
    ktrc_fill_header(nHdrP, hookword, 0, _TR_FORMAT_X, -1);

    va_start(vargs, fmt);
    stringLen = vsnprintf(tBuf+sizeof(newHdr_t),
                          LXTRACE_MAX_DATA-sizeof(newHdr_t), fmt, vargs) + 1;
    va_end(vargs);

    if (stringLen > LXTRACE_MAX_DATA-sizeof(newHdr_t))
    {
      Kprint("_XTrace: argument too long.  len=%d max=%u hook=0x%X\n",
             stringLen, (unsigned int)(LXTRACE_MAX_DATA-sizeof(newHdr_t)-1), hookword);
      stringLen = LXTRACE_MAX_DATA-sizeof(newHdr_t);
    }

    /* Patch the string and record lengths now that vsnprintf has calculated
       the length that it formatted */
    nHdrP->hdr.trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
    trLength = sizeof(*nHdrP) + nHdrP->hdr.trSLen;
    checkTODTrace();
    appendKernelTrace((newHdr_t*)tBuf, trLength);
  }
  else
#endif /* SHARED_TRC_BUF */
  {
#ifdef LXTRACE_USE_RELAY
    tBuf[LXTRACE_MAX_DATA]='\0';
    if (isRLEnabled())
    {
      hdrP = (trc_datahdr_t*)tBuf;
      ktrc_fill_dataheader(hdrP, hookword, 0, _TR_FORMAT_X, -1);
    
      va_start(vargs, fmt);
      stringLen = vsnprintf(tBuf+sizeof(trc_datahdr_t),
                            LXTRACE_MAX_DATA-sizeof(trc_datahdr_t), fmt, vargs) + 1;
      va_end(vargs);
    
      if (stringLen > LXTRACE_MAX_DATA-sizeof(trc_datahdr_t))
      {
        Kprint("_XTrace: argument too long.  len=%d max=%u hook=0x%X\n",
               stringLen, (unsigned int)(LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)-1), hookword);
        stringLen = LXTRACE_MAX_DATA-sizeof(trc_datahdr_t);
      }
    
      /* Patch the string and record lengths now that vsnprintf has calculated
         the length that it formatted */
      hdrP->trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
      trLength = sizeof(hdr) + hdrP->trSLen;
    
      rl_trc_write(tBuf, trLength, 0);
      return;
    }
#endif /* LXTRACE_USE_RELAY */
    /* Build a data header and append it to the trace file.  Since the length
       is not yet known, use the maximum.  It will be patched to the correct
       value later. */
    ktrc_fill_dataheader(&hdr, hookword, 0, _TR_FORMAT_X, -1);
    rc = trc_start_record(&hdr, LXTRACE_MAX_DATA, false);
    /* If the header was successfully written, format the string directly
       into the trace buffer */
    if (rc == 0)
    {  
      va_start(vargs, fmt);
      stringLen = vsnprintf(lxthe.writeBuf.nextP,
                            LXTRACE_MAX_DATA-sizeof(trc_datahdr_t), fmt, vargs) + 1;
      va_end(vargs);
      if (stringLen > LXTRACE_MAX_DATA-sizeof(trc_datahdr_t))
      {
        Kprint("_XTrace: argument too long.  len=%d max=%u hook=0x%X\n",
               stringLen, (unsigned int)(LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)-1), hookword);
        stringLen = LXTRACE_MAX_DATA-sizeof(trc_datahdr_t);
      }
  
      /* Patch the string and record lengths now that vsnprintf has calculated
         the length that it formatted */
      lxthe.hdrP->trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
      lxthe.tHdrP->trLength = sizeof(hdr) + lxthe.hdrP->trSLen;
  
      /* Advance pointer into trace buffer by the length of the string just
         appended */
      lxthe.writeBuf.nextP += lxthe.hdrP->trSLen;
  
      /* Trace record complete */
      trc_end_record();
    }
  }
}
#endif /* GPFS_PRINTF */
