/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
/* @(#)67       1.35.1.39  src/avs/fs/mmfs/ts/kernext/gpl-linux/tracelin.c, mmfs, avs_rhrz, rhrzs022a 11/17/14 08:32:12 */

/**************************************************************************
 *
 * Loadable kernel module that implements the trace device. 
 *
 **************************************************************************/


#ifndef GPFS_PRINTF

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef KBUILD_MODNAME
#define KBUILD_MODNAME tracedev
#endif

/* If trace is built into kernel, pick up GPFS flag definitions from a file
   rather than requiring them to be defined on the command line. */
#ifndef MODULE
/* #include <linux/ktrace.h> */
#endif

#include <Shark-gpl.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/string.h>

#if LINUX_KERNEL_VERSION > 2060900
#include <linux/hardirq.h> /* in_interrupt */
#else
#include <asm/hardirq.h> /* in_interrupt */
#endif
#include <asm/uaccess.h> /* copy_to/from_user */


#include <stdarg.h>
#include <Trace.h>
#include <lxtrace.h>
#include <verdep.h>

#ifdef SHARED_TRC_BUFF
#include <cxiAtomic.h>
#endif

#ifdef LXTRACE_USE_RELAY
extern int rl_trc_write(const char *bufP, size_t nBytes, int nb);
extern int rl_init(void);
extern void rl_cleanup(void);
extern void rl_trc_stop(void);
spinlock_t seqLock;
#endif
volatile unsigned long long seq;

#ifdef __64BIT__
# define Int64 long long
# define ARGLEN 8
# define ARGTYPE Int64
#else
# define Int32 int
# define ARGLEN 4
# define ARGTYPE Int32
#endif /* __64BIT__ */

#if LINUX_KERNEL_VERSION > 2060900 || \
  defined(GPFS_ARCH_PPC64) || defined(GPFS_ARCH_X86_64)
#define EXPORTKDUMPDEV
#endif

#ifdef EXPORTKDUMPDEV
static int major_kdump = -1;
#endif

#if defined(MODULE)
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION ("GPFS portability layer (tracing module)");
MODULE_AUTHOR ("IBM <gpfs@us.ibm.com>");
#if (LINUX_KERNEL_VERSION >= 2060500)
MODULE_INFO (supported, "external");
#endif
#endif /* MODULE */

#ifdef SHARED_TRC_BUFF
static char *traceBufferP = NULL; /* For the shared memory buffer */
static Int64 trcBufferSize = 0;
int *traceBufferOffsetP;
#endif /* SHARED_TRC_BUF */
/* If trace is built into kernel, then this is a dummy module */
#ifndef KTRACE

/* the daemon's task structure (for signal) */
static struct task_struct *taskP;

/* The writeLock serializes trace writers.  It should be avoided by
 * other operations in order to allow the writers to continue unimpeded.
 * The writeLock must be held when accessing the following fields in the
 * trace header element:  nWaits, nBuffers, nLost, writeBuf
 */
static spinlock_t writeLock;

/* The readLock serializes trace operations, as well as most other access
 * to the trace header element.  Whenever both readLock and writeLock are
 * required, readLock is always to be acquired first.
 */
static struct semaphore readLock;

/* The readFull flag synchronizes access to readBuf by readers and writers.
   Writers set this after filling readBuf and wait for this to be clear
   before filling readBuf.  Readers use this flag to tell if readBuf has
   any data and clear this after processing.  Using an atomic variable
   allows steady-state tracing to be done without readers needing to
   acquire a lock that would block writers.  Note that atomic operations
   do not generally act as memory barriers, so explicit barrier calls may
   be necessary before or after accessing readFull.  Spinlocks act as
   partial memory barriers, so explicit barriers can be avoided in some
   cases where spinlocks are used. */
static atomic_t readFull;

/* Trace Header Element - THE anchor for the trace state */
trcdev_header_t   lxthe;
static wait_queue_head_t daemonWaitQ;
static int mode = LXTRACE_BLK;

/* Export pointers to internal data structures for debugging */
struct
{
  trcdev_header_t   *lxtheP;
  wait_queue_head_t *daemonWaitQP;
  struct semaphore  *readLockP;
  spinlock_t        *writeLockP;
} TraceVarAddrs = { &lxthe, &daemonWaitQ, &readLock, &writeLock };

/* A trcdev_buffer is dirty if there is any data in it (nextP != beginP) AND the
 * dirtyP has not yet been advanced (by trc_read) past the data (to nextP) */
#define IS_DIRTY(b) (b.nextP != b.beginP && b.nextP != b.dirtyP)

/* A trace record passed from a user thread consists of a data header
   followed by the marshalled arguments */
struct trcRec
{
#ifdef SHARED_TRC_BUFF
    newtrc_header_t h0;
    trc_datahdr_t hdr;
    char data[LXTRACE_MAX_DATA-sizeof(trc_datahdr_t) - sizeof(newtrc_header_t)];
#else
    trc_datahdr_t hdr;
    char data[LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)];
#endif /* SHARED_TRC_BUF */
};

#ifdef SHARED_TRC_BUFF
/* If we need to copy the data into the shared trace buffer from the daemon
   using an ioctl we duplicate the code from ktrccalls.c */

/* Get the current clock cycle count. Duplicated in ktrcalls.c */
static inline UInt64 __readCycleCounter()
{
#if defined(GPFS_ARCH_X86_64)
  UInt32 low, high;
  __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high));
  return (low | ((UInt64)(high) << 32));
#elif defined(GPFS_ARCH_PPC64) || defined(GPFS_ARCH_POWER)
  UInt64 val;
  __asm__ __volatile__("mftb %0" : "=r" (val));
  return val;
#elif defined(GPFS_ARCH_S390X)
  UInt64 val;
  __asm__ __volatile__("stck 0(%1)" : "=m" (val) : "a" (&val) : "cc");
  return val;
#else
  #error "Can't get CPU count for this linux architecture"
#endif
}
/* rdtscll is defined in various other places in the kernel,
   so use a different name here only to quiet compiler warning. */
#define rdtscall(val) \
  ((val) = __readCycleCounter())

#define ROUNDUP(X,Y) ( (Y) * ( ((X)+(Y)-1) / (Y) ) )

/* Copy trace record into trace buffer.
   Duplicated from ktrccalls.c. */
static void tracelinAppendKernelTrace(void* tBuf, int trRecLen)
{
    char* traceP;
    int copyLen;
    int reserveLen;
    Int64 oldOffset;
    Int64 newOffset;
    Boolean csSuccess;
    Int64 bufRemaining;
    int i;
    Int64 *KTraceOffsetPtr = (Int64 *)traceBufferP;
    int *KTraceWrapPtr = (int *)(traceBufferP + sizeof(Int64));
    char *KTraceBufferPtr = traceBufferP + CACHE_LINE;
    Int64 KTraceBufferSize = trcBufferSize - CACHE_LINE;
    newHdr_t* newHdrP = (newHdr_t *)tBuf;


    /* Round trace record size up to a cache line boundary */
    copyLen = trRecLen;
    reserveLen = ROUNDUP(copyLen, CACHE_LINE);
  /* Obtain a slot in the trace buffer for the current record. This
     assignment is done outside the loop since ATOMIC_SWAP automatically
     updates oldOffset if the swap fails. */
  oldOffset = *KTraceOffsetPtr;
  do
  {
    newOffset = oldOffset + reserveLen;
    bufRemaining = KTraceBufferSize - newOffset;
    if (bufRemaining <= 0)
      newOffset -= KTraceBufferSize;
    csSuccess = ATOMIC_SWAP64((atomic_p)KTraceOffsetPtr,
                              &oldOffset, newOffset);
  } while (!csSuccess);

    /* Fill in the timestamp and other header fields */
  newHdrP->h0.trMagic = LXTRACE_SWAP_MAGIC;
  rdtscall(newHdrP->h0.trTime);
  newHdrP->h0.trProcess = (pid_t) current->pid;
  newHdrP->h0.trCPU = smp_processor_id();
  newHdrP->h0.trLength = trRecLen;
  newHdrP->h0.trWrapPoint = 0;
  newHdrP->h0.trWrapPoint = *KTraceWrapPtr;

  /* Copy the trace record into the slot just obtained.  It may wrap
     around the trace buffer. */
  bufRemaining = KTraceBufferSize - oldOffset;
  traceP = KTraceBufferPtr + oldOffset;
  if (copyLen <= bufRemaining)
  {
    memcpy(traceP, newHdrP, copyLen);
  }
  else
  {
    memcpy(traceP, newHdrP, bufRemaining);
    traceP = KTraceBufferPtr;
    memcpy(traceP, (char*)newHdrP + bufRemaining, copyLen-bufRemaining);
  }

  /* If we are wrapping or made exactly to the end, update the wrap counter so the next entry
     will know it has wrapped.
     Note that some threads that came after us may already have created trace record using the
     old counter before we got to increment or vise-versa, some threads that came before us but
     got heldup uptil now would be using the incremented counter. This is acceptable risk as
     correcting that would require atomicity in setting TraceOffset and wrap counter. Lxtrace will
     identify records in such category. */
  if (copyLen >= bufRemaining)
    *KTraceWrapPtr = *KTraceWrapPtr + 1;
}

/* To get useful Time of Day (TOD) traces, there must be
   a reasonable number of them. Make sure there are enough in each wrap.
   We have matching code in gpfslinux.C, ktrccalls.c and tracelin.c (for when we
   are copying traces into the shared buffer via ioctl). No synchronization is
   done to track how many TOD timestamps we are emitting, so each trace path
   will emit TODPerWrap TOD traces per wrap. */
int KprevWrapCount = -1;
#define  KTODPerWrap 4
 int KTODThisWrap = 0; 
 int *KTODThisWrapP = &KTODThisWrap; 

/* Checks if it's time to emit a TOD trace; if so, emit one.
   Note that this function is currently unused a*/
static void checkTODTrace()
{
  struct timeval trTime;
  newHdr_t *hdrP;
  char tBuf[LXTRACE_MAX_DATA+1];
  long long sec, usec;
  int csSuccess;
  Int64 *KTraceOffsetPtr = (Int64 *)traceBufferP;
  int *KTraceWrapPtr = (int *)(traceBufferP + sizeof(Int64));
  Int64 KTraceBufferSize = trcBufferSize - CACHE_LINE;
  long long endStamp = LXTRACE_TOD_END_MAGIC;

  if (KprevWrapCount != *KTraceWrapPtr)
    KTODThisWrap = 0;

  /* To make sure we don't emit too many TOD traces,
     force it to use compare and swap to track the number of
     emitted TOD traces. */
  do
  {
    if (*KTraceOffsetPtr < KTODThisWrap * (KTraceBufferSize / KTODPerWrap))
    {
      return;
    }
    csSuccess = ATOMIC_SWAP((atomic_p)KTODThisWrapP,
                            &KTODThisWrap, KTODThisWrap + 1);
  } while (!csSuccess);

  KTODThisWrap++;
  KprevWrapCount = *KTraceWrapPtr;
  memset(tBuf, '9', LXTRACE_MAX_DATA);
  hdrP = (newHdr_t*)tBuf;
  hdrP->hdr.trHook = LXTRACE_TOD_MAGIC;
  hdrP->hdr.trNArgs = 2;
  hdrP->hdr.trSPos = _TR_FORMAT_I;
  hdrP->hdr.trSLen = 0;

  do_gettimeofday(&trTime);
  sec = (long long)trTime.tv_sec;
  usec = (long long)trTime.tv_usec;
  memcpy(tBuf+sizeof(newHdr_t), &(sec), sizeof(long long));
  memcpy(tBuf+sizeof(newHdr_t)+ sizeof(long long), &(usec), sizeof(long long));
  memcpy(tBuf+sizeof(newHdr_t)+ 2*sizeof(long long), &(endStamp), sizeof(long long));
  tracelinAppendKernelTrace((newHdr_t*)tBuf, sizeof(newHdr_t) + 3*sizeof(long long));
  printk("appending kernel tod trace (in tracelin) from %lld.%lld\n",sec, usec);
}
#endif /* SHARED_TRC_BUFF */


/* Updating state information requires the writeLock in addition to
 * readLock.  The readLock is widely held where the trace header is 
 * manipulated, but for the brief period of updating the state field, get the 
 * writeLock as well.
 */
static void setTraceState(trcdev_state_t newState)
{
  spin_lock(&writeLock);
  lxthe.state = newState;
  wake_up(&daemonWaitQ);
  spin_unlock(&writeLock);
}


/* Return true if the specified hookword is currently being traced. 
   Hookwords are no longer used, so this just checks if trace is active. */
inline Boolean isTraced(uint hw)
{
  #ifdef SHARED_TRC_BUFF
  return (lxthe.state == trc_active || lxthe.state == trc_shared);
#else
  return lxthe.state == trc_active;
#endif
}

//free allocated buffer pages
static void free_buf(void* buf)
{
  #ifdef SHARED_TRC_BUFF
  vfree(buf);
#else
  unsigned int i;
  vunmap(buf);

  for (i = 0; i < lxthe.page_count; i++)
    __free_page(lxthe.page_array[i]);

  vfree(lxthe.page_array);
#endif /* SHARED_TRC_BUFF */
}

//allocate buffer in pages to avoid 64MB limitation of vmalloc on SLES9
static void* realloc_buf(unsigned long size, void *buf)
{
  void *mem;
  unsigned int i, j, n_pages;
  struct page **page_array;

  #ifdef SHARED_TRC_BUFF
  mem = vmalloc(size);
  return mem;
  #else

  size = PAGE_ALIGN(size);
  n_pages = size >> PAGE_SHIFT;

  //vmalloc on SLES9 use kmalloc here
  //so the page number can not exceed 128K/sizeof(struct page *)
  page_array = vmalloc(n_pages*sizeof(struct page *));
  if (!page_array)
    return NULL;

  for (i = 0; i < n_pages; i++) {
    page_array[i] = alloc_page(GFP_KERNEL);
    if (unlikely(!page_array[i]))
      goto err;
  }

  mem = vmap(page_array, n_pages, VM_MAP, PAGE_KERNEL);
  if (!mem)
    goto err;

  //free previous buffers
  if (buf)
  {
    free_buf(buf);
  }

  lxthe.page_array = page_array;
  lxthe.page_count = n_pages;

  return mem;

err:
  for (j = 0; j < i; j++)
    __free_page(page_array[j]);
  vfree(page_array);
  
  return NULL;
#endif /* SHARED_TRC_BUFF */
}

static void *alloc_buf(unsigned long size)
{
  return realloc_buf(size, NULL);
}


/* Construct the static trace header element ("lxthe"). 
 * trc_open will allocate buffers and set the appropriate values. */
static void trc_init()
{
  spin_lock_init(&writeLock);
  sema_init(&readLock, 1);

  /* Note:  Locks are not needed here.  There better not be any other
     threads trying to access lxthe at this point.  If there were, then
     what would happen if a thread tried to acquire the locks a few
     instructions earlier, before we initialized the locks? */

  lxthe.major = 0; /* dynamic assignment (by register_chrdev in trc_register) */
  lxthe.minor = 0;
  lxthe.bufSize = 0;
  lxthe.nOpens = 0;
  lxthe.nWaits = 0;
  lxthe.nBuffers = 0;
  lxthe.nLost = 0;
  atomic_set(&readFull, 0);
#ifdef LXTRACE_USE_RELAY
  spin_lock_init(&seqLock);
#endif
  seq = 0;
  taskP = NULL;
  init_waitqueue_head(&daemonWaitQ);

  lxthe.writeBuf.beginP = NULL;
  lxthe.writeBuf.endP = NULL;
  lxthe.writeBuf.nextP = NULL;
  lxthe.writeBuf.dirtyP = NULL;
  lxthe.writeBuf.lastP = NULL;

  lxthe.readBuf = lxthe.writeBuf;

  lxthe.state = trc_initialized;
}

/* Destroy the static trace header element (lxthe) */
static void trc_term()
{
  /* Note: Locks are not needed here.  We're about to re-initialize, so
     if anyone is still using lxthe at this point, we would clobber them. */

  /* The two buffers are allocated together.  Free them both here. */
  if (lxthe.writeBuf.beginP)
    free_buf(MIN(lxthe.writeBuf.beginP, lxthe.readBuf.beginP));
#ifdef SHARED_TRC_BUFF
  if(traceBufferP != NULL)
  {
    free_buf(traceBufferP);
    traceBufferP = NULL;
    printk("Freeing shared trace buffer\n");
    trcBufferSize = 0;
  }
#endif
  /* (re)initialize all fields.  Rather than copy all the stuff that happens 
   * in trc_init, we can use it here to reset all the fields. */
  trc_init();
}

#ifdef EXPORTKDUMPDEV
#ifdef GPFS_ARCH_IA64
static int kdump_get_stack (struct unw_frame_info *info, unsigned long *ipList)
{
  unsigned long ip, sp, bsp;
  int num = 0;

  do {
    unw_get_ip(info, &ip);
    if ((ip == 0) || (num > 255))
      break;

    ipList[num++] = ip;
  } while (unw_unwind(info) >= 0);

  return num;
}
#endif

/* On ia64, kdump_read can read backtrace for every thread and copy back to user */
static ssize_t kdump_read(struct file *fileP, char *bufP, size_t nBytes, loff_t *ppos)
{
  int rc= -EINVAL;
  unsigned long kaddr = (unsigned long)*ppos;

#if (defined(GPFS_ARCH_X86_64) && LINUX_KERNEL_VERSION >= 2061600 \
    && (defined(SUSE_LINUX)|| defined(REDHAT_AS_LINUX) \
        || defined(FEDORA_LINUX) || defined(DEBIAN_LINUX) || defined(UBUNTU_LINUX))) || \
    ((defined(GPFS_ARCH_PPC64) || defined(GPFS_ARCH_IA64)) && \
    LINUX_KERNEL_VERSION >= 2061100)
  /* rw_verify_area does not allow kernel addr range,
     so a read() will fail with EINVAL.
     We subtracted the base kernel addr in kdump.c and add back in here. */
  unsigned long highBits = GPFS_KERNEL_OFFSET;
#else
  unsigned long highBits = 0;
#endif

#ifdef GPFS_ARCH_IA64
  unsigned long taddr = 0;
  unsigned long ipList[256];
  int num = 0;
  struct task_struct *task = NULL;
  struct unw_frame_info info;

  if (nBytes == 0)
  {
    /* read backtrace for specified thread */
    task = (struct task_struct*)((unsigned long)*ppos + highBits);
    if (!task)
      return rc;
    else
    {
      unw_init_from_blocked_task(&info, task);
      num = kdump_get_stack(&info, ipList);
    }

    if (num == 0)
      rc = num;
    if (num > 0)
      if (copy_to_user(bufP, ipList, sizeof(unsigned long) * num) == 0)
        rc = num;
    return rc;
  }
#endif

  kaddr += highBits;

#if (defined(GPFS_ARCH_X86_64) && LINUX_KERNEL_VERSION >= 2061800 \
     && defined(REDHAT_AS_LINUX))
 
  /* Since RHEL 5, definition of macro __pa(x) has been changed. 
   * #define __pa(x)                 ((unsigned long)(x) - PAGE_OFFSET)
   * This definition is ok for i386 where kernel text/data is part of lineraly
   * mapped region. But that is not case with x86_64 where 40 MB
   * (ffffffff80000000 - ffffffff82800000) virtual address space has been fixed
   * for kernel text mapping assuming kernel load address is phy 0.
   * Therefore address need to be readjust accordingly
   */

  kaddr = (kaddr >= __START_KERNEL_map) ? 
          (kaddr -__START_KERNEL_map + PAGE_OFFSET) : kaddr ;
#endif    

#if ((! defined(GPFS_ARCH_PPC64)) && (! defined(GPFS_ARCH_IA64)))
  /* on ppc64, virt_addr_valid only can check kernel address belongs to directly
   * mapping part (PAGE_OFFSET - HIGH_MEMORY), it consider adress in vm area as
   * in-valid
   */
  if (virt_addr_valid(kaddr))
#endif
    if (copy_to_user(bufP, (void *)((unsigned long)*ppos + highBits), nBytes)==0)
      rc=nBytes;
  return((ssize_t)rc);
}

static int kdump_open(struct inode *inodeP, struct file *fileP)
{
   try_module_get(THIS_MODULE);
   fileP->f_pos=0;
   return 0;
}

static int kdump_close(struct inode *inodeP, struct file *fileP)
{
   module_put(THIS_MODULE);
   return 0;
}

static loff_t kdump_lseek(struct file *fileP, loff_t offset, int orgin)
{
   loff_t rc;

   if (orgin != 0)
      return(-EAGAIN);

   fileP->f_pos = offset;

   return(offset);
}
#endif

/* The device open operation.  The first open is initiated by the trace daemon,
 * and comes after registration.  It results in the allocation of the trace
 * buffers, and identifying the trace daemon (so it can be signalled when
 * buffers are ready to be read).  */
static int trc_open(struct inode *inodeP, struct file *fileP)
{
  int rc = 0;

  /* Serialize multiple opens and prevent state changes */
  down(&readLock);
  /* Only the daemon opens the device O_RDWR, and only does so when turning
   * trace on.
   */
  if ((fileP->f_flags & O_ACCMODE) == O_RDWR)
  {
     if (lxthe.state != trc_initialized)
     {
       #ifdef SHARED_TRC_BUFF
       /* For the shared trace buffer, multiple processes must open the
          trace device RDWR so they can mmap it into userspace */
       rc = 0;
       try_module_get(THIS_MODULE);
       lxthe.nOpens += 1;
       #else
       rc = -EALREADY;
       #endif /* SHARED_TRC_BUF */
       goto exit;
     }

    /* The first open (lxtrace on) requires initialization of the header. */
    lxthe.minor = MINOR(inodeP->i_rdev);

    /* Only supporting one such device */
    if (lxthe.minor > 0)
    {
      rc = -ENODEV;
      goto exit;
    }

    /* If not configured otherwise, use the default buffer size. */
    if (lxthe.bufSize == 0)
      lxthe.bufSize = DEF_TRC_BUFSIZE;

#ifndef SHARED_TRC_BUFF
    /* Allocate dual trace buffers (new records go into the write buffer,
     * and the daemon reads (via trc_read) from the read buffer). */
    lxthe.writeBuf.beginP = alloc_buf(2*lxthe.bufSize);
    if (!lxthe.writeBuf.beginP)
    {
      rc = -ENOMEM;
      goto exit;
    }
    lxthe.writeBuf.endP = lxthe.writeBuf.beginP + lxthe.bufSize - 1;
    lxthe.writeBuf.nextP = lxthe.writeBuf.beginP;
    lxthe.writeBuf.dirtyP = lxthe.writeBuf.beginP;
    lxthe.writeBuf.lastP = lxthe.writeBuf.beginP;
    lxthe.writeBuf.bufNum = 1;

    lxthe.readBuf.beginP = lxthe.writeBuf.beginP + lxthe.bufSize;
    lxthe.readBuf.endP = lxthe.readBuf.beginP + lxthe.bufSize - 1;
    lxthe.readBuf.nextP = lxthe.readBuf.beginP;
    lxthe.readBuf.dirtyP = lxthe.readBuf.beginP;
    lxthe.readBuf.lastP = lxthe.readBuf.beginP;
    lxthe.readBuf.bufNum = 2;

    /* With the Shared Trace Buffer more processes will be opening the trace device
       with O_RDWR to use mmap. Use an ioctl to set taskP instead */
    /* Save pointer to the daemon task information, and mark the
     * device open. */
    taskP = current;
#endif
    setTraceState(trc_opened);

    /* Since threads that handle VM page-outs also do traces, set flag so
       that we will not get blocked waiting to allocate pages.  Otherwise a
       deadlock could occur if the page-out thread was waiting for us to
       empty the trace buffer, and we are waiting for the page-out thread
       to free some pages. */
    current->flags |= PF_MEMALLOC;
  }

  /* Applications must open the trace device O_WRONLY.  These opens do not
   * require any processing.  If the daemon has turned tracing on, the open
   * is allowed and subsequent write() calls will be handled.  If the daemon
   * has NOT turned tracing on, the application open will be granted, but
   * subsequent write() calls will NOOP
   * until the daemon turns trace on (state == trac_active). */

  else if ((fileP->f_flags & O_ACCMODE) != O_WRONLY)
  {
    /* After "trace on", subsequent trace control commands open O_RDONLY.  */
#ifdef SHARED_TRC_BUFF
    if (lxthe.state != trc_active && lxthe.state != trc_shared)
#else
    if (lxthe.state != trc_active)
#endif
    {
      rc = -EALREADY;
      goto exit;
    }
  }

  lxthe.nOpens += 1;

  try_module_get(THIS_MODULE);

exit:
  up(&readLock);
  return rc;
}

/* The device read operation.  This is to be used only by the trace daemon to
 * retrieve trace buffers for the purposes of writing to the output file. */
static ssize_t trc_read(struct file *fileP, char *bufP, size_t nBytes, loff_t *ppos)
{
  ssize_t nDone = 0;
  ssize_t nReady;

  /* All access to lxthe.readBuf is protected via the readLock.  */
  down(&readLock);

  /* Only the trace daemon is allowed to read. */
  if (taskP && taskP->pid != current->pid)
  {
    nDone = -EPERM;
    goto exit;
  }

  /* See if there is data waiting to be processed by the daemon.  Read is
   * allowed here during normal operation (trc_active) and as trace is
   * terminating (this to get the last group of buffered records). */
  if ((lxthe.state == trc_active || lxthe.state == trc_stopped) && 
      atomic_read(&readFull))
  {
    /* Be sure that we don't access readBuf until after readFull is set */
    rmb();

    if (IS_DIRTY(lxthe.readBuf))
    {
      /* Make sure that the caller's buffer is large enough to hold 
       * what we have. */
      nReady = lxthe.readBuf.nextP - lxthe.readBuf.beginP;
      if (nReady > nBytes)
      {
        printk("trc_read: state %d ready data %d exceed buffer size %d! "
               "redBuf: nextP %p beginP %p lastP %p, writeBuf: nextP %p beginP %p lastP %p\n",
               lxthe.state, (int)nReady, (int)nBytes, 
               lxthe.readBuf.nextP, lxthe.readBuf.beginP, lxthe.readBuf.lastP,
               lxthe.writeBuf.nextP, lxthe.writeBuf.beginP, lxthe.writeBuf.lastP);
        nDone = -EFBIG;
        goto exit;
      }

      if (copy_to_user(bufP, lxthe.readBuf.dirtyP, nReady))
      {
        nDone = -EFAULT;
        goto exit;
      }

      nDone = nReady;
      lxthe.readBuf.dirtyP += nDone;
    }
    
    /* Allow writers to use readBuf */
    if (!IS_DIRTY(lxthe.readBuf))
    {
      wmb();
      atomic_set(&readFull, 0);
      wake_up(&daemonWaitQ);
    }
  }

exit:
  up(&readLock);
  return nDone;
}

static void my_send_sig_info(int mySig, struct siginfo * sigData, struct task_struct *taskP)
{
   struct task_struct *g, *tsP;
#if defined(HAS_TASKLIST_LOCK)
   TASK_READ_LOCK;
   do_each_thread(g,tsP)
   {
     if (tsP == taskP)
     {
       send_sig_info(mySig, sigData, tsP);
       break;
     }
   } while_each_thread(g,tsP);
   TASK_READ_UNLOCK;
#else
   rcu_read_lock();
   if(taskP == NULL)
   {
     printk("ERROR: attempt to send signal to uninitialized pid.\n");
     rcu_read_unlock();
     return;
   }
   tsP = FIND_TASK_BY_PID(taskP->pid);
   if(tsP == NULL)
   {
      rcu_read_unlock();
      return;
   }
   if (tsP->pid == taskP->pid)
     send_sig_info(mySig, sigData, tsP);
   rcu_read_unlock();
#endif
}

/* Internal routine to schedule i/o of the trace buffer.
   NOTE that this routine is called while holding the writeLock. */
static void trc_signal_io()
{
  trcdev_buffer_t tBuf;
  struct siginfo sigData;

  /* DBGASSERT(atomic_read(&readFull) == 0); */

  /* Switch the buffers.  We don't have to worry about trc_read looking at
     readBuf while we do this because it always verify that readFull is
     non-zero before accessing readBuf. */
  rmb();
  tBuf = lxthe.readBuf;
  lxthe.readBuf = lxthe.writeBuf;
  lxthe.writeBuf= tBuf;
  lxthe.nBuffers++;             /* Number of buffers filled */

  /* Mark readBuf full so that writers won't switch to it until after the
     daemon has processed it.  Do write memory barrier to ensure that
     our change to readBuf makes it to memory before readFull is set. */
  wmb();
  atomic_set(&readFull, 1);

  // record the position of last time
  // so that if this buffer is not totally overwritten, 
  // we can still get some previous trace data
  if( mode == LXTRACE_NB )
    lxthe.writeBuf.lastP = lxthe.writeBuf.nextP;
  /* Reset the (new) writeBuf to a clean state */
  lxthe.writeBuf.dirtyP = lxthe.writeBuf.nextP = lxthe.writeBuf.beginP;

  /* Debug for 471707: Since all trace records begin with a header the 
   * very first thing in a dirty buffer should be a valid header.  If 
   * this is not the case, print debug information to the log file.
   */
  if (IS_DIRTY(tBuf) && (((trc_header_t *)tBuf.beginP)->trMagic != LXTRACE_SEQ_MAGIC))
  {
    printk("trc_signal_io: bad trace buffer! trMagic 0x%X\n",
          ((trc_header_t *)tBuf.beginP)->trMagic);
    printk("trc_signal_io: begin 0x%p end 0x%p next 0x%p dirty 0x%p isDirty %d\n",
           (trc_header_t *)tBuf.beginP, (trc_header_t *)tBuf.endP, 
           (trc_header_t *)tBuf.nextP, (trc_header_t *)tBuf.dirtyP, 
           IS_DIRTY(tBuf));
  }
  // only flush trace data in blocking mode or when trace stop in overwrite mode
  if( mode == LXTRACE_BLK || lxthe.state == trc_stopped )
  {
    /* Signal daemon that there is a trace buffer to be read and processed. */
    sigData.si_signo = SIGIO;
    sigData.si_errno = 0;
    sigData.si_code  = SI_KERNEL;
    my_send_sig_info(SIGIO, &sigData, taskP);
  }
}


/* Return true if trace writer will have to wait for daemon to make room
   for a trace record of the specified length. */
static int writeMustBlock(int len)
{
  if( mode == LXTRACE_BLK )
    return (len > lxthe.writeBuf.endP - lxthe.writeBuf.nextP + 1 &&
            atomic_read(&readFull) && lxthe.state == trc_active);
  else
    return 0;
}


/* Reserves space for a trace record whose data header plus arguments
   totals nBytes.  Returns 0 if space was reserved for the trace record,
   or non-zero if space could not be found because the buffer is full
   and cantBlock is set, or because tracing is not enabled.  If space
   was reserved successfully, builds a trace header, then copies the
   trace header and the given data header into the trace device buffer.
   After returning from this routine, trc_append_record should be called
   0 or more times, then trc_end_record.  Returns with writeLock held iff
   0 was returned. */
int
trc_start_record(trc_datahdr_t * hdrP, size_t nBytes, int cantBlock)
{
  int trclen;
  trc_header_t tHdr;

  /* Construct the trace record header */
  tHdr.trMagic = LXTRACE_SEQ_MAGIC;

  /* Wait to set the timestamp (tHdr.trTime) until after all serialization.
   * When multiple threads call trace, they don't necessarily get the
   * writeLock in FIFO order so setting the timestamp here can result
   * in times going backwards in the trace file.
   *   do_gettimeofday(&tHdr.trTime);
   */

  tHdr.trProcess = current->pid;
  tHdr.trCPU = smp_processor_id();
  tHdr.trLength = nBytes;
  trclen = nBytes + sizeof(trc_header_t);

  /* Serialize access to writeBuf */
  spin_lock(&writeLock);

  /* If this trace record will not fit in the write buffer, and the read
     buffer is still full, and trace is active, then we must wait for the
     daemon to empty the read buffer. */
  if (writeMustBlock(trclen))
  {
    if (cantBlock)
    {
      lxthe.nLost++;
      spin_unlock(&writeLock);
      return 1;
    }

    if (lxthe.state != trc_active)
    {
      spin_unlock(&writeLock);
      return 1;
    }

    lxthe.nWaits++;
    do
    {
      spin_unlock(&writeLock);
      wait_event(daemonWaitQ, !writeMustBlock(trclen));
      spin_lock(&writeLock);
    } while (writeMustBlock(trclen));
  }
  
  if (lxthe.state != trc_active)
  {
    spin_unlock(&writeLock);
    return 1;
  }

  /* Will the trace record fit into the write buffer?  If not, then we can
     swap with the read buffer which must be empty at this point (else we
     wouldn't have come out of previous wait loop). */
  if (trclen > lxthe.writeBuf.endP - lxthe.writeBuf.nextP + 1)
  {
    /* Swap write buffer with read buffer and signal daemon to process the
       data. */
    trc_signal_io();

    /* This could be an assert, since write buffer must be empty now. */
    if (trclen > lxthe.writeBuf.endP - lxthe.writeBuf.nextP + 1)
    {
      spin_unlock(&writeLock);
      return 1;
    }
  }

  //The seq variable are protected by spinlock
  tHdr.trSeq = ++seq;

  /* Now that there isn't anything to block the writing of this
   * record, insert the timestamp.
   */
  do_gettimeofday(&tHdr.trTime);

  /* Insert the header stamp into the buffer ahead of the application
     record and remember its location. */
  lxthe.tHdrP = (trc_header_t *)lxthe.writeBuf.nextP;
  tHdr.trBuf = lxthe.writeBuf.bufNum;
  //no sequence number in header
  memcpy(lxthe.writeBuf.nextP, &tHdr, sizeof(trc_header_t));
  lxthe.writeBuf.nextP += sizeof(trc_header_t);

  /* Move the application trace header directly into the trace buffer and
     remember its location */
  lxthe.hdrP = (trc_datahdr_t *)lxthe.writeBuf.nextP;
  memcpy(lxthe.writeBuf.nextP, hdrP, sizeof(*hdrP));
  lxthe.writeBuf.nextP += sizeof(*hdrP);

  /* Return with writeLock still held */
  return 0;
}


/* Append a portion of a trace record to the write buffer.  Must have
   previously called trc_start_record. */
void trc_append_record(const void* bufP, size_t nBytes)
{
  /* Move the application trace record directly into the trace buffer */
  memcpy(lxthe.writeBuf.nextP, bufP, nBytes);
  lxthe.writeBuf.nextP += nBytes;
}


/* Finish a trace record */
void trc_end_record()
{
  spin_unlock(&writeLock);
}

/* internal function used both for trc_write and trc_ioctl_write */
static inline int trc_write_internal(struct trcRec tr, size_t nBytes)
{
  int dataBytes;
  int rc;

#ifdef LXTRACE_USE_RELAY
  /* If relay enabled, use it, else  try the old style trace. */
  if (isTraced(tr.hdr.trHook))
  {
    if(rl_trc_write((const char *)&tr.hdr,nBytes,0)>=0)
      return nBytes;
  }
#endif

  /* The beginning of the trace record is a hookword number.  Verify that
     the specified hookword is being traced.  If not, return as if the trace
     was successful. */
  if (isTraced(tr.hdr.trHook))
  {
    rc = trc_start_record(&tr.hdr, nBytes, false);
    if (rc == 0)
    {
      dataBytes = nBytes - sizeof(tr.hdr);
      if (dataBytes > 0)
        trc_append_record(&tr.data[0], dataBytes);
      trc_end_record();
    }
  }
  return nBytes;
}
static ssize_t
trc_write(struct file *fileP, const char *bufP, size_t nBytes, loff_t *posP)
{
  struct trcRec tr;
  int rc;

  /* A special value of nBytes == 0 means the caller only wants to
     find out whether tracing is enabled. */
  if (nBytes == 0 && (lxthe.state != trc_active))
    return -EBADF;

  /* Copy trace record from user address space */
  if (nBytes < 4 || nBytes > LXTRACE_MAX_DATA)
    return -EINVAL;
  if (copy_from_user(&tr, bufP, nBytes))
    return -EFAULT;

  return trc_write_internal(tr, nBytes);
}

/* Before close, a sync of the trace device will flush the records
 * still in the read buffer (even though it might not be full).  A
 * close without this call could result in the loss of these records.
 * Must not call fsync from daemon termination signal handler because
 * that could deadlock if a SIGIO is still pending.
 */
static int
/* struct file *fP, struct dentry *dP, loff_t start, loff_t end, int datasync */
trc_fsync_internal(FILE_OPS_FSYNC_PARAMETERS)
{
  struct siginfo sigData;
  spin_lock(&writeLock);
  //for overwrite mode, only flush data when trace is stopped
  if ( mode == LXTRACE_NB )
  {
    if( lxthe.state != trc_stopped )
    {
      spin_unlock(&writeLock);
      return 0;
    }
    else if(atomic_read(&readFull))
    {

      /* Signal daemon that there is a trace buffer to be read and processed. */
      sigData.si_signo = SIGIO;
      sigData.si_errno = 0;
      sigData.si_code  = SI_KERNEL;
      my_send_sig_info(SIGIO, &sigData, taskP);
    }
  }
  /* If read buffer is still full, wait for daemon to process it */
  while (atomic_read(&readFull) &&
         (lxthe.state == trc_active || lxthe.state == trc_stopped))
  {
    spin_unlock(&writeLock);
    wait_event(daemonWaitQ,
               !(atomic_read(&readFull) &&
                 (lxthe.state == trc_active || lxthe.state == trc_stopped)));
    spin_lock(&writeLock);
  }
  
  /* Allow fsync during normal operation OR after ioctl(trc_end) has
     disabled further trace writing (allows an fsync before close to
     flush the buffered records). */
  if (lxthe.writeBuf.nextP != lxthe.writeBuf.beginP &&
      (lxthe.state == trc_active || lxthe.state == trc_stopped))
  {
    //for the last buffer of overwrite mode
    //we can still get part of the previous trace data which is not overwritten 
    if (lxthe.writeBuf.nextP < lxthe.writeBuf.lastP )
      lxthe.writeBuf.nextP = lxthe.writeBuf.lastP;
    trc_signal_io();
  }
  spin_unlock(&writeLock);
  return 0;
}


/* The externally visible version of trc_fsync_internal */
int trc_fsync()
{
  TRC_FSYNC_IMPLEMENTATION();
}

/* The device close operation. */
static int trc_close(struct inode *inodeP, struct file *fileP)
{
  down(&readLock);

  /* The trace daemon only closes the device upon termination. */
  if (taskP && taskP->tgid == current->tgid)
  {
#ifdef LXTRACE_USE_RELAY
    //the trace daemon can not handle SIGKILL
    //trace must be stoppped here 
    rl_trc_stop();
#endif
    /* The final trace daemon close.  Reset for subsequent use. */
    setTraceState(trc_initialized);

    /* We don't really need writeLock here since writers won't do anything
       after state is set to trc_initialized, but it doesn't hurt. */
    spin_lock(&writeLock);
    lxthe.nWaits = 0;
    lxthe.nBuffers = 0;
    lxthe.nLost = 0;
    spin_unlock(&writeLock);
    taskP = NULL;
    current->flags &= ~PF_MEMALLOC;

    /* Free the two trace buffers. */
    if (lxthe.writeBuf.beginP)
    {
      //vfree(MIN(lxthe.writeBuf.beginP, lxthe.readBuf.beginP));
      free_buf(MIN(lxthe.writeBuf.beginP, lxthe.readBuf.beginP));

      lxthe.writeBuf.beginP = NULL;
      lxthe.writeBuf.endP = NULL;
      lxthe.writeBuf.nextP = NULL;
      lxthe.writeBuf.dirtyP = NULL;
      lxthe.writeBuf.lastP = NULL;

      lxthe.readBuf = lxthe.writeBuf;
    }
    seq = 0;
  }

  lxthe.nOpens -= 1;

  module_put(THIS_MODULE);

  up(&readLock);
  return 0;
}

static void terminateOldDaemon()
{
  struct siginfo sigData;
  sigData.si_signo = SIGTERM;
  sigData.si_errno = 0;
  sigData.si_code  = SI_KERNEL;
  my_send_sig_info(SIGTERM, &sigData, taskP);
}

/* ioctl op used to for low-level access to trace operation. */
static long trc_ioctl_internal(struct file *fileP,
                     unsigned int op, unsigned long kx_args)
{
  int h, rc = 0;
  Boolean readLockHeld = false;
  struct kArgs args_cp;
  struct kArgs *args = (struct kArgs *)kx_args;
  char *p;
  char *newBufP;
  char *trc_dumpP;
  char *trc_nextP;
  struct siginfo sigData;
  int waitCount = 0;
  struct trcRec tr;
  int buffSize;
#ifdef SHARED_TRC_BUFF
  Int64 daemonTrcBufSize;
  int i;
  struct newHdr_t* newHdrP;
struct cxiMemoryMapping_t
{
  char *vaddr;        /* daemon address mapping */
#ifdef SSEG_SWIZZLE_PTRS
  char *kvaddr;       /* kernel address of memory area (shared segment only) */
#endif
  int kBytes;         /* size of the area in kilobytes */
  int vindex;         /* index in shared segment mapping array */
  cxiXmem_t xmemDesc; /* cross-memory descriptor */
};
  struct cxiMemoryMapping_t userspaceMM;

  UIntPtr vmallocStart, vmallocEnd;
  unsigned long ssegUserMin, ssegUserMax;
  char *localTrcBufferP;
  int *tmpPtr;
  int tscSpeed = -1;
#endif /* SHARED_TRC_BUF */
  if(op != trc_ioctl_write)
  {
    down(&readLock);
    readLockHeld = true;
  }

  switch (op)
  {
    case trc_begin:
      if (lxthe.state == trc_active)
      {
        rc = -EALREADY;
        break;
      }
      if (lxthe.state != trc_opened)
      {
        rc = -EBADF;
        break;
      }
      /* get the argument array */
      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }
      mode = args_cp.arg1;
      setTraceState(trc_active);
      break;
    case trc_end:
#ifdef SHARED_TRC_BUFF
      if (lxthe.state == trc_shared)
      {
        /* For blocking mode, we set the state to 'stopped' then
           when the lxtrace daemon exits we set it to 'initialized'
           so tracing can be restarted correctly. Just set it to initialized
           immediately for overwrite mode. */
        setTraceState(trc_initialized);
        /* printk("Freeing shared trace buffer and disabling traces\n"); */
        /* if(traceBufferP == NULL) */
        /* { */
        /*   printk("WARNING. ATTEMPTED TO FREE PREVIOUSLY FREED TRACE BUFFER.\n"); */
        /*   break; */
        /* } */
        /* localTrcBufferP = traceBufferP; */
        /* traceBufferP = NULL; */
        /* free_buf(localTrcBufferP); */
        /* trcBufferSize = 0; */
      }
      else
#endif
      if (lxthe.state != trc_active)
        rc = -EBADF;
      else
      {
        setTraceState(trc_stopped);
        up(&readLock);
        readLockHeld = false;
        trc_fsync();

        /* Signal the daemon to terminate. */
        sigData.si_signo = SIGTERM;
        sigData.si_errno = 0;
        sigData.si_code  = SI_KERNEL;
        my_send_sig_info(SIGTERM, &sigData, taskP);
      }

      /* Wait for lxtrace to terminate, but don't wait forever.
         At this point the signal has been delivered to lxtrace,
         but it may take some time for the process to exit.  Since
         lxthe.state is changed from trc_stopped to trc_initialized
         in trc_close(), which is called when lxtrace exits, if we
         return control to the caller right away, there'd be a window
         when tracing has ostensibly been stopped, and it should be
         OK to start tracing again, but trying to do so would fail
         with EALREADY in trc_open because lxthe.state is not what
         the code expects.  So we give lxtrace some time to terminate.
         Something could go seriously wrong, and lxtrace may get stuck,
         we don't wait forever. */
      while (lxthe.state == trc_stopped && waitCount++ < 10)
      {
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(100);
      }
      break;

    case trc_recycle:
#ifdef SHARED_TRC_BUFF
      if (lxthe.state == trc_shared)
        rc = -EBUSY;
      else
#endif /* SHARED_TRC_BUFF */
      if (lxthe.state != trc_active)
        rc = -EBADF;
      else
      {
        /* Signal the daemon to recycle. */
        sigData.si_signo = SIGIO;
        sigData.si_errno = 0;
        sigData.si_code  = SI_KERNEL;
        my_send_sig_info(SIGIO, &sigData, taskP);
      }
      break;
#ifdef SHARED_TRC_BUFF
      /* Return the locaTion in kernel space of the trace buffer */
    case trc_alloc_query:
      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }
      if (traceBufferP == NULL)
      {
        printk("shared trace buffer is not allocated\n");
        rc = -ENOMEM;
        break;
      }
      if( copy_from_user(&userspaceMM, (struct cxiMemoryMapping_t *)args_cp.arg2,
                         sizeof(struct cxiMemoryMapping_t)))
      {
        rc = -EFAULT;
        break;
      }
      userspaceMM.kvaddr = (char *) traceBufferP;
      if(copy_to_user((struct cxiMemoryMapping_t *)args_cp.arg2,
                       &userspaceMM, sizeof(struct cxiMemoryMapping_t)))
         rc = -EFAULT;
      if(copy_to_user((Int64 *)args_cp.arg1, &trcBufferSize,
                      sizeof(Int64 *)))
        rc = -EFAULT;
      if(copy_to_user((char *)args_cp.arg4, &traceBufferP,
                      sizeof(char *)))
        rc = -EFAULT;
#if defined(GPFS_ARCH_X86_64) && LINUX_KERNEL_VERSION >= 2062200
      if (tsc_khz > 10000 && tsc_khz < 100000000)
        tscSpeed = tsc_khz;
      else
        tscSpeed = -1;
#elif defined(GPFS_ARCH_PPC64) || defined(GPFS_ARCH_POWER) || defined(GPFS_ARCH_S390X) || LINUX_KERNEL_VERSION < 2062200
      tscSpeed = -1;
#else
  #error "Can't determine tsc clock rate for this linux architecture"
#endif
      if(copy_to_user((int *)args_cp.arg5, &tscSpeed,
                      sizeof(int)))
        rc = -EFAULT;

      printk("Daemon obtained pointer to trace buffer of size %lld\n",   trcBufferSize);
      break;
    case trc_size_query:
      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }
      if (copy_to_user((Int64 *)args_cp.arg1, &trcBufferSize,
                    sizeof(Int64)) ||
          copy_to_user((Int64 *)args_cp.arg2, &trcBufferSize,
                       sizeof(Int64)))
      {
        rc = -EFAULT;
        break;
      }
      /* We also return the current buffer offset for lxtrace.
         If the trace buffer is not yet full and has never wrapped,
         report lxtrace can use that information to not write out the
         empty data in the buffer. Round that to the nearest 4K for DIO */
      if(traceBufferP != NULL && trcBufferSize > 0)
      {
        Int64 blocksize = 4096;
        Int64 alignedOffset = (Int64)*((Int64 *)traceBufferP);
        if(alignedOffset % blocksize != 0)
          alignedOffset += blocksize - alignedOffset % blocksize;
        if(copy_to_user((Int64 *)args_cp.arg2, &alignedOffset,
                        sizeof(Int64 )))
          rc = -EFAULT;
      }
      break;
    case trc_dealloc:
      printk("Freeing shared trace buffer\n");
      if(traceBufferP == NULL)
      {
        printk("WARNING. ATTEMPTED TO FREE PREVIOUSLY FREED TRACE BUFFER.\n");
        break;
      }
      localTrcBufferP = traceBufferP;
      traceBufferP = NULL;
      free_buf(localTrcBufferP);
      trcBufferSize = 0;
      break;
    case trc_set_lxtrace_pid:
    /* Save pointer to the daemon task information */
      taskP = current;
      break;
    case trc_alloc:
      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }
      if (traceBufferP != NULL)
      {
        printk("The buffer is already allocated\n");
        setTraceState(trc_shared);
        break;
      }
      if (copy_from_user(&daemonTrcBufSize, (Int64 *)args_cp.arg1, sizeof(Int64)))
      {
        rc = -EFAULT;
        break;
      }
      if(daemonTrcBufSize <= 0)
      {
        printk("Invalid argument to trc_alloc; must request positive size\n");
        rc = -EFAULT;
        break;
      }
      trcBufferSize = daemonTrcBufSize;
      trcBufferSize += PAGE_SIZE - (trcBufferSize % PAGE_SIZE);
      traceBufferP = (char *) alloc_buf(trcBufferSize);
      if (traceBufferP == NULL)
      {
        printk("Memory allocation for shared buffer failed\n");
        rc = -EFAULT;
        break;
      }
      memset(traceBufferP, 0, trcBufferSize);
      printk("Allocated trace buffer of size %lld (rounded up to PAGESIZE from %lld) at 0x%lX\n", trcBufferSize,
             daemonTrcBufSize, (long)traceBufferP);
      if (lxthe.state == trc_active && taskP != NULL)
      {
        /* If we had previously been running in blocking mode
           kill the old lxtrace daemon if necessary */
        terminateOldDaemon();
        taskP = NULL;
      }
      setTraceState(trc_shared);

      break;

#endif /* SHARED_TRC_BUF */

    case trc_bufSize:

      /* The daemon may call ioctl to change the desired buffer size.
         On open, buffers of the default size are allocated.  This call
         frees the current buffers (replacing them with new ones).  Any
         trace records currently in the buffers will be lost. */

      if (lxthe.state != trc_opened)
      {
        rc = -EPERM;
        break;
      }

      /* get the argument array */
      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }
      
      /* Allocate the new (dual) trace buffers.
       * arg1 is the requested buffer size */
      //newBufP = vmalloc(2*args_cp.arg1);
      newBufP = (char *)realloc_buf(2*args_cp.arg1, MIN(lxthe.writeBuf.beginP, lxthe.readBuf.beginP));
      if (!newBufP)
      {
        rc = -ENOMEM;
        break;
      }

      /* Free the previous buffers.  Since the state is currently
       * "trc_opened" and we are holding readLock, neither readers nor
       * writers can be using the buffers at this time. */
      //if (lxthe.writeBuf.beginP)
        //vfree(MIN(lxthe.writeBuf.beginP, lxthe.readBuf.beginP));

      lxthe.bufSize = args_cp.arg1;
      lxthe.writeBuf.beginP = newBufP;

      lxthe.writeBuf.endP = lxthe.writeBuf.beginP + lxthe.bufSize - 1;
      lxthe.writeBuf.nextP = lxthe.writeBuf.beginP;
      lxthe.writeBuf.dirtyP = lxthe.writeBuf.beginP;
      lxthe.writeBuf.lastP = lxthe.writeBuf.beginP;

      lxthe.readBuf.beginP = lxthe.writeBuf.beginP + lxthe.bufSize;
      lxthe.readBuf.endP = lxthe.readBuf.beginP + lxthe.bufSize - 1;
      lxthe.readBuf.nextP = lxthe.readBuf.beginP;
      lxthe.readBuf.dirtyP = lxthe.readBuf.beginP;
      lxthe.readBuf.lastP = lxthe.readBuf.beginP;
      break;

    case trc_dump:

      /* format trace header information and return to daemon */
      trc_dumpP = vmalloc(LXTRACE_DUMP_SIZE);
      if (trc_dumpP == NULL)
      {
        rc = -ENOMEM;
        break;
      }

      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }

      /* Block writers so that we can look at writeBuf. */
      spin_lock(&writeLock);

      /* Format the state information suitable for displaying by 
       * the daemon. 
       */
      trc_nextP = trc_dumpP;
      sprintf(trc_nextP, "Trace Header Element: 0x%8p\n", &lxthe);
      trc_nextP += strlen(trc_nextP);

      /* Global information on device number, buffer sizes, 
       * and lost records. 
       */
      sprintf(trc_nextP, "  Major %d Minor %d bufSize 0x%X nOpens %d "
                         "nBuffers %d nLost %d nWaits %d Daemon %d Mode %d\n",
                         lxthe.major, lxthe.minor, lxthe.bufSize, lxthe.nOpens,
                         lxthe.nBuffers, lxthe.nLost, lxthe.nWaits,
                         taskP ? taskP->pid: 0, mode);
      trc_nextP += strlen(trc_nextP);

      sprintf(trc_nextP, "\n");
      trc_nextP += strlen(trc_nextP);

      /* Append buffer information */
      sprintf(trc_nextP, "  writeBuf: beginP 0x%p endP 0x%p nextP 0x%p "
                         "dirtyP 0x%p isDirty %d\n",
                         lxthe.writeBuf.beginP, lxthe.writeBuf.endP,
                         lxthe.writeBuf.nextP, lxthe.writeBuf.dirtyP, 
                         IS_DIRTY(lxthe.writeBuf));
      trc_nextP += strlen(trc_nextP);

      sprintf(trc_nextP, "  readBuf : beginP 0x%p endP 0x%p nextP 0x%p "
                         "dirtyP 0x%p isDirty %d\n",
                         lxthe.readBuf.beginP, lxthe.readBuf.endP,
                         lxthe.readBuf.nextP, lxthe.readBuf.dirtyP, 
                         IS_DIRTY(lxthe.readBuf));
      trc_nextP += strlen(trc_nextP);

#if 0
      /* verify dumpBuf size */
      sprintf(trc_nextP, "  dumpBuf size %d (used %d)\n",
              LXTRACE_DUMP_SIZE, (trc_nextP-trc_dumpP));
      trc_nextP += strlen(trc_nextP);
#endif 
#ifdef SHARED_TRC_BUFF
      if (traceBufferP != NULL && lxthe.state == trc_shared)
      {
        sprintf(trc_nextP, "  trcOffset: %lld bufferSize: %lld wrap: %d tsc speed %d\n",
                (Int64)*((Int64 *)traceBufferP), trcBufferSize,
                (int)*((int *)(traceBufferP + sizeof(Int64))),
                (int)*((int *)(traceBufferP + sizeof(Int64) + sizeof(int))));
        trc_nextP += strlen(trc_nextP);
      }
      sprintf(trc_nextP, " lxtrace pid %d trace state %d\n", taskP ? taskP->pid : 0, lxthe.state);
#endif
      spin_unlock(&writeLock);

      /* arg1 is the user buffer size, arg2 is the address of the buffer */
      if (copy_to_user((char *)args_cp.arg2, trc_dumpP,
                       MIN(strlen(trc_dumpP)+1, args_cp.arg1)))
        rc = -EFAULT;

      vfree(trc_dumpP);
      break;

    case trc_ioctl_write:
#ifdef SHARED_TRC_BUFF
      if (lxthe.state != trc_active && lxthe.state != trc_shared)
#else
      if (lxthe.state != trc_active)
#endif
      {
        rc = -EBADF;
        break;
      }
      //used for performance reason, provide same functionality as trc_write
      if (copy_from_user(&args_cp, args, sizeof(args_cp)))
      {
        rc = -EFAULT;
        break;
      }

      /* Copy trace record from user address space */
      if (args_cp.arg1 < 4 || args_cp.arg1 > LXTRACE_MAX_DATA)
      {
        rc = -EINVAL;
        break;
      }
      #ifdef SHARED_TRC_BUFF
      /* During daemon startup or shutdown we might
         pass in data using ioctls while in overwrite mode. */
      if (copy_from_user(&tr, (char *)args_cp.arg2, args_cp.arg1))
      {
        rc = -EFAULT;
        break;
      }
      if (lxthe.state == trc_shared)
      {
        //        checkTODTrace();
        tracelinAppendKernelTrace((newHdr_t *)&tr, args_cp.arg1);
        break;
      }
      else
      {
        /* In blocking mode, make sure we don't try to pass the
           newtrc_header_t along. */
        buffSize = args_cp.arg1 - sizeof(newtrc_header_t);
      }
#else
      if (copy_from_user(&tr, (char *)args_cp.arg2, args_cp.arg1))
      {
        rc = -EFAULT;
        break;
      }
      buffSize = args_cp.arg1;
#endif /* SHARED_TRC_BUFF */
      rc = trc_write_internal(tr, buffSize);
      break;

    default:
      rc = -EINVAL;
      break;
  }

  if (readLockHeld)
    up(&readLock);

  return rc;
}

#ifdef SHARED_TRC_BUFF
int trc_mmap(struct file *file, struct vm_area_struct *vma)
{
  /* Maps the trace buffer that was vmalloc'ed by
     ioctl trc_alloc to userspace (either for the daemon or for lxtrace).
     Userspace will call mmap and this function handles that mapping. */
  Int64 mmapLength = vma->vm_end - vma->vm_start;
  unsigned long pfn, start = vma->vm_start;
  int rc;
  char *ptrP  = (char *)traceBufferP;
  if(traceBufferP == NULL)
    return -ENOMEM;
  if(lxthe.state != trc_shared)
  {
    printk("Attempt to mmap buffer when not in overwrite mode\n");
    return -ENOMEM;
  }
  /* This shouldn't ever happen, but just in case things get out of sync */
  if(trcBufferSize != mmapLength)
    {
      printk("mmapLength %d, trcBufferSize %lld. Mismatch, mmap fails\n", (int)mmapLength, trcBufferSize);
      return -EINVAL;
    }
  for (mmapLength; mmapLength > 0; mmapLength -= PAGE_SIZE)
  {
    pfn = vmalloc_to_pfn(ptrP);
#ifdef GPFS_ARCH_S390X
    rc = remap_pfn_range(vma, start, pfn, PAGE_SIZE, PAGE_KERNEL);
#else
    rc = remap_pfn_range(vma, start, pfn, PAGE_SIZE, PAGE_SHARED);
#endif
    if(rc < 0)
    {
      printk("Error in trc_mmap, remap_pfn_range rc %d\n",rc);
      return rc;
    }
    start += PAGE_SIZE;
    ptrP += PAGE_SIZE;
  }
  vma->vm_flags |= VM_RESERVED;
  return 0;
}
#endif /* SHARED_TRC_BUF */


#ifdef HAVE_UNLOCKED_IOCTL
static long trc_unlocked_ioctl(struct file *fileP, unsigned int op,
                                 unsigned long kx_args)
{
  return trc_ioctl_internal(fileP,op,kx_args);
}
#endif
static int trc_ioctl(struct inode *inodeP, struct file *fileP,
                     unsigned int op, unsigned long kx_args)
{
  return trc_ioctl_internal(fileP,op,kx_args);
}

static struct file_operations trc_ops =
{
  llseek:     NULL,
  read:       trc_read,  /* read op allows the daemon to retrieve records */
  write:      trc_write, /* Trace points write to the device */
#ifdef HAS_READDIR
  readdir:    NULL,
#endif
  poll:       NULL,
#ifdef HAVE_UNLOCKED_IOCTL
  unlocked_ioctl: trc_unlocked_ioctl,
#else
  ioctl:      trc_ioctl, /* control op to change buffering or dump state */
#endif
#ifdef SHARED_TRC_BUFF
  mmap:       trc_mmap,  /* Maps the kernel space trace buffer to user space */
#endif /* SHARED_TRC_BUF */
  open:       trc_open,  /* Prepare the device for tracing */
  flush:      NULL,
  release:    trc_close, /* Terminate tracing and close the device */
  fsync:      trc_fsync_internal, /* Sync all buffered data to the daemon */
  fasync:     NULL,
  lock:       NULL,
};

#ifdef EXPORTKDUMPDEV
static struct file_operations kdump_ops =
{
  llseek:     kdump_lseek,
  read:       kdump_read,  /* read op allows the daemon to retrieve records */
  write:      NULL, /* Trace points write to the device */
#ifdef HAS_READDIR
  readdir:    NULL,
#endif
  poll:       NULL,
  mmap:       NULL,
  open:       kdump_open,  /* Prepare the device for tracing */
  flush:      NULL,
  release:    kdump_close, /* Terminate tracing and close the device */
  fsync:      NULL, /* Sync all buffered data to the daemon */
  fasync:     NULL,
  lock:       NULL,
};
#endif
/* Register the trace device "/dev/trace" and save the major number in 
 * the header 
 */
static int trc_register()
{
  int major = register_chrdev(0, "trace", &trc_ops);
  if (major < 0)
    return major;
  lxthe.major = major;
#ifdef EXPORTKDUMPDEV
  major_kdump = register_chrdev(0, "kdump", &kdump_ops);
#endif
  return 0;
}

/* Unregister the trace device */
static void trc_unregister()
{
  unregister_chrdev(lxthe.major, "trace");
  lxthe.major = 0;
#ifdef EXPORTKDUMPDEV
  if (major_kdump >= 0)
     unregister_chrdev(major_kdump, "kdump");
  major_kdump = 0;
#endif

}

/* Module initialization */
static int __init myModuleInit(void)
{
  int rc;
  trc_init();

#ifdef LXTRACE_USE_RELAY
  rc = rl_init();
  if( rc < 0 )
    return rc;
#endif

  return trc_register();
}

static void __exit myModuleExit(void)
{
#ifdef LXTRACE_USE_RELAY
  rl_cleanup();
#endif
  trc_unregister();
  trc_term();
}

module_init(myModuleInit);
module_exit(myModuleExit);

#endif /* GPFS_PRINTF */
#endif /* KTRACE */
