/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
/* @(#)35	1.17.1.7  src/avs/fs/mmfs/ts/kernext/ibm-kxi/lxtrace.h, mmfs, avs_rhrz, rhrzs018a 3/11/14 17:18:28 */

#ifndef _h_lxtrace
#define _h_lxtrace

/*
 * Interfaces to the Linux kernel trace device
 *
 */

#ifndef _LINUX_TIME_H
#include <sys/time.h>
#endif

/* Size of the ioctl trc_dump output buffer */
#define LXTRACE_DUMP_SIZE 768

#define LXTRACE_MAX_DATA 512
#define LXTRACE_MAGIC 0xACE98271
#define LXTRACE_COMP_MAGIC 0xACE98272
#define LXTRACE_SEQ_MAGIC 0xACE98273
#ifdef SHARED_TRC_BUFF
#define LXTRACE_SWAP_MAGIC 0xACE98274
#define LXTRACE_TOD_MAGIC 0xACE98275
#define LXTRACE_TOD_END_MAGIC 0xACE98276
#endif /* SHARED_TRC_BUF */
#define LXTRACE_NODATA_MAGIC 0x0

#define LXTRACE_NR_CPUS 256
/* Number of "$n" substitutions supported in a single trace record */
#define LXTRACE_MAX_FORMAT_SUBS 20

#define RL_DISABLE 0x0
#define LXTRACE_BLK 0x1
#define LXTRACE_NB 0x2
#define RL_FLUSH 0x4
#define RL_NB_PAUSE 0x8
 /* While running in NB mode, we will briefly enable BLK mode to copy the data out. We also
    set this flag so we know that this is temporary */

#define RL_BUF_GLOBAL 0x1
#define RL_BUF_PERCPU 0x2
#define RL_BUF_CTIME  0x4

#define DEBUG_DIR "/sys/kernel/debug/gpfs"

#define LXTRACE_PROJID 0xff

#define LXTRACE_MTYPE 0x1

#define LXTRACE_DIO_ALIGN 4096
#define LXTRACE_DIO_MASK (~(LXTRACE_DIO_ALIGN-1)) 
//#define LXTRACE_DEBUG
/* Convenience macros for printing error and debug messages */
#if defined(GPFS_LINUX)
#define DP(fmt,args...) fprintf(stderr, fmt, ## args)
#define DFP(fmt,args...) if (lxtraceDebug) fprintf(stderr, fmt, ## args)
#elif defined(GPFS_SOLARIS) || defined(GPFS_WINDOWS)
#define DP(fmt, ...) fprintf(stderr, fmt, __VA_ARGS__)
#ifdef LXTRACE_DEBUG
#define DFP(fmt,...) fprintf(stderr, fmt, __VA_ARGS__)
#else
#define DFP(fmt, ...) ((void)0)
#endif
typedef unsigned int Boolean;
#define false 0
#define true 1
#endif /* GPFS_... */


/* Header of a trace record to be passed to the trace device */
typedef struct
{
    uint trHook;          /* Trace hook word that uniquely identifies the
                              -                           format of this trace record */
    char trNArgs;         /* Number of integer arguments.  Each argument is as
                           long as the word width of the machine.  Does not
                           count string, generic, or float arguments. */
    unsigned char trSPos; /* Position of string argument within argument list.
                           May also take on special values to indicate all
                           integers or all integers followed by a float. */
    short trSLen;         /* If format is string or generic, contains the length
                           of the variable part, rounded up to a multiple
                           of the word size */
} trc_datahdr_t;
/* followed by char arg[LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)] */

/* Definitions of the trace record format */
typedef struct
{
  uint trMagic;         /* Magic number to allow resynch after trace
                           wraparound */
  struct timeval trTime; /* Timestamp of trace record */
  pid_t trProcess;      /* ID of process making the trace */
#ifdef GPFS_SOLARIS
  uint64_t trThread;
#endif
  unsigned char trBuf;  /* bufNum of buffer used for trace */
  unsigned char trCPU;  /* CPU number of process making the trace */
  short trLength;       /* Number of bytes that follow, consisting of a
                           trc_datahdr_t followed by the application
                           data */
} trc_noseq_header_t;
typedef struct
{
  uint trMagic;         /* Magic number to allow resynch after trace
                           wraparound */
  struct timeval trTime; /* Timestamp of trace record */
  pid_t trProcess;      /* ID of process making the trace */
#ifdef GPFS_SOLARIS
  uint64_t trThread;
#endif
  unsigned char trBuf;  /* bufNum of buffer used for trace */
  unsigned char trCPU;  /* CPU number of process making the trace */
  short trLength;       /* Number of bytes that follow, consisting of a
                           trc_datahdr_t followed by the application
                           data */
  unsigned long long trSeq;            /* sequence number */
} trc_header_t;
typedef struct
{
  uint trMagic;         /* Magic number to allow resynch after trace
                           wraparound */
  int dataLength;       /* original data length */
  int compLength;       /* compressed data length */
} trc_comp_header_t;

/* Followed by the application trace data "char data[Length]" */

#ifdef SHARED_TRC_BUFF
typedef struct
{
  UInt32 trMagic;     /* Magic number to allow resynch after trace
                           wraparound */
  UInt32 trProcess;      /* ID of process making the trace */
  UInt64 trTime;        /* CPU cycle count of trace */

  Int32 trWrapPoint; /* How many times the trace has wrapped */
  char unused1; /* For padding */
  unsigned char trCPU;  /* CPU number of process making the trace */
  short trLength;       /* Number of bytes that follow, consisting of a
                           trc_datahdr_t followed by the application
                           data */
} newtrc_header_t;

/* While the code doesn't make this obvious in the kernel/daemon,
   when lxtrace comes to format the data it will be of the form
   "newHdr_t" followed by a char array of the amount given in newtrc_header_t */
typedef struct
{
  newtrc_header_t h0;
  trc_datahdr_t hdr;
} newHdr_t;
/* followed by char arg[LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)] */
#endif /* SHARED_TRC_BUF */


/* The following device operations are supported for the trace device:
    trc_open     Prepare the device for tracing
    trc_ioctl    Device control operation (see the trace_op definition)
    trc_read     Allows the daemon to retrieve the trace records
    trc_write    Trace records are writen to the device
    trc_fsync    Sync all buffered data to the daemon
    trc_close    Terminate tracing and close the device
 */

/* Device ioctl operations */
enum trace_op
{
  trc_begin,    /* Enable tracing for the specified trace types */
  trc_end,      /* Fence new writes in preparation for close */
  trc_bufSize,  /* Change the device buffer size */
  trc_dump,      /* Dump trace state information */
#ifdef SHARED_TRC_BUFF
  trc_alloc,     /* Allocate a new shared trace buffer  */
  trc_dealloc,   /* Deallocate the shared trace buffer */
  trc_alloc_query,     /* Makes sure lxtrace has allocated the shared buffer and
                           returns its location in kernel space */
  trc_size_query, /* Checks how large of a trace buffer has been allocated */
  trc_set_lxtrace_pid, /* Sets the pid for lxtrace so it can be signalled later */
  
#endif /* SHARED_TRC_BUF */
#ifdef GPFS_SOLARIS
  trc_fsynch,   /* Solaris has fsync as an ioctl */
#endif
  trc_recycle,
  trc_ioctl_write
};

/* Structure for passing parameters to the trace device through ioctls */
#ifdef GPFS_SOLARIS
struct kArgs
{
  uint64_t arg1;
  uint64_t arg2;
  uint64_t arg3;
  uint64_t arg4;
  uint64_t arg5;
};
#else /* Linux */
struct kArgs
{
  long arg1;
  long arg2;
  long arg3;
  long arg4;
  long arg5;
};
#endif

#ifdef __cplusplus
extern "C"
{
#endif
  int trc_fsync();
#ifdef __cplusplus
}
#endif


/* Allowable file sizes for the lxtrace command */
#define MIN_TRC_FILESIZE 1*1024*1024
#define DEF_TRC_FILESIZE 128*1024*1024
#define MAX_TRC_FILESIZE 16*1024*1024*1024LL

/* Allowable buffer sizes for the lxtrace command.  Two buffers of this
   size will be allocated. */
#define MIN_TRC_BUFSIZE 1*1024*1024 
#define DEF_TRC_BUFSIZE 4*1024*1024
#define MAX_TRC_BUFSIZE 64*1024*1024

/* Allowable total buffer size for relay */
#define MIN_TRC_OVERWRITE_BUFSIZE 16*1024*1024
#define DEF_TRC_OVERWRITE_BUFSIZE 128*1024*1024

/* Allowable per-cpu buffer number for relay */
#define MIN_TRC_BUFNUM 2
#define DEF_TRC_BUFNUM 4
#define MAX_TRC_BUFNUM 32

/* Allowable per-cpu buffer number for relay */
#define MIN_TRC_COMPLEVEL 0
#define DEF_TRC_COMPLEVEL 6
#define MAX_TRC_COMPLEVEL 9

// Default flush interval for relay 
#define DEF_TRC_INT 0 

/* Trace device name */
#ifdef GPFS_SOLARIS
#define TRC_DEVICE "/devices/pseudo/gpfs@1:tracedev"
#else
#define TRC_DEVICE "/dev/trace0"
#endif

#define LXTRACE_CMD_PATH "/usr/lpp/mmfs/bin/lxtrace"

#define LXTRACE_MSG_LEN 16
typedef struct trc_msgbuf
{
  long mtype;      /* message type, must be > 0 */
  char mtext[LXTRACE_MSG_LEN];  /* message data */
} trc_msg;
#define LXTRACE_INIT_OK "success"
#define LXTRACE_INIT_ERR "error"

#define TRC_MAX_ID2NAME 128
typedef struct { char s[TRC_MAX_ID2NAME]; } idname_t;
#ifdef GPFS_WINDOWS
#define TRCFMT_NAME "tstrcfmt"
#else
#define TRCFMT_NAME "lxtrace format"
#endif

/* id2name conversion specifier. */
typedef struct i2nSpec_t
{
  int argNum;       // number of the argument to be converted
  char *i2nNameP;   // conversion function name
} i2nSpec_t;


#ifdef __KERNEL__
/* Trace device structures */
#ifdef SHARED_TRC_BUFF
static void appendKernelTrace(newHdr_t* newHdrP, int trRecLen);
#endif /* SHARED_TRC_BUF */

/* Possible states of the trace device */
typedef enum trcdev_state_t
{
  trc_initialized,  /* header constructed */
  trc_opened,       /* device opened      */
  trc_active,       /* ready for tracing  */
  trc_stopped,       /* tracing disabled   */
  #ifdef SHARED_TRC_BUFF
  trc_shared,       /* Running in shared buffer mode */
  #endif
} trcdev_state_t;

typedef struct trcdev_buffer_t
{
  char * beginP; /* first byte of the buffer */
  char * endP;   /* last byte of the buffer */
  char * nextP;  /* offset of next record to be read/written */
  char * dirtyP; /* offset of next byte to be read */
  char * lastP;  /* offset of last nextP */
  int bufNum;    /* buffer number */
} trcdev_buffer_t;

typedef struct trcdev_header_t
{
  int major;                /* major number of the trace device */
  int minor;                /* minor number of the trace device */
  int bufSize;              /* number of bytes in the trace buffers */
  int nOpens;               /* number of times the device is open */
  int nWaits;               /* number of times we had to wait for the daemon */
  int nBuffers;             /* number of buffers filled with trace data */
  int nLost;                /* number of traces lost due to full buffer */
  trcdev_state_t state;     /* status of the trace device */
  trcdev_buffer_t writeBuf; /* The trace buffer being written */
  trcdev_buffer_t readBuf;  /* The buffer being read by the daemon */
  trc_header_t * tHdrP;     /* address of trc_header_t for the record
                               currently being appended */
  trc_datahdr_t * hdrP;     /* address of trc_datahdr_t for the record
                               currently being appended */
  struct page **page_array; /* array of allocated buffer pages */
  unsigned int page_count;  /* number of allocated buffer pages */ 
} trcdev_header_t;

#if defined(LXTRACE_WANT_RELAY) && LINUX_KERNEL_VERSION >= 2061400
#define LXTRACE_USE_RELAY
#endif
#else
#if defined(LXTRACE_WANT_RELAY) 
#define LXTRACE_USE_RELAY
#endif
#endif /* __KERNEL __ */

#endif /* _h_lxtrace */
